import os
import unittest
import numpy as np
import scipy

from snudda.neurons.neuron_morphology_extended import NeuronMorphologyExtended


class NeuronMorphologyExtendedTestCase(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        #swc_file = os.path.join(os.path.dirname(__file__), "validation", "ballanddoublestick", "double.swc")

        swc_file = os.path.join(os.path.dirname(__file__), "validation", "striatum", "fs",
                                "str-fs-e161205_FS1-mMTC180800A-IDB-v20190312", "MTC180800A-IDB-cor-rep.swc")
        self.nm = NeuronMorphologyExtended(swc_filename=swc_file, load_morphology=True)  #, use_cache=False

    def test_input_location(self, stage="dist_to_soma"):

        synapse_density = "(d > 100e-6)*1"
        rng = np.random.default_rng(123456)  #

        xyz, sec_id, sec_x, dist_to_soma = self.nm.dendrite_input_locations(synapse_density_str=synapse_density,
                                                                            rng=rng, num_locations=100)

        # Please note that num_locations currently does not guarantee 100 synpases when requesting it

        # 3e-6 due to compartment length sampled at 3 micrometers
        self.assertTrue((dist_to_soma > 100e-6 - 3e-6).all())

        self.assertEqual(xyz.shape[1], 3)
        self.assertEqual(len(dist_to_soma), xyz.shape[0])
        self.assertEqual(len(dist_to_soma), len(sec_id))
        self.assertEqual(len(sec_id), len(sec_x))

        # Repeat test but for smaller than 100
        synapse_density = "(d < 200e-6)*1"

        xyz, sec_id, sec_x, dist_to_soma = self.nm.dendrite_input_locations(synapse_density_str=synapse_density,
                                                                            rng=rng, num_locations=100)
        # 3e-6 due to compartment length sampled at 3 micrometers
        self.assertTrue((dist_to_soma < 200e-6 + 3e-6).all())

#   -- rand_rotation is now moved to rotation.py
#
#     def test_rand_rotation(self, stage="rand_rotation"):
#
#         for idx in range(0, 100):
#             rot_mat = self.nm.rand_rotation_matrix()
#             self.assertAlmostEqual(np.linalg.det(rot_mat), 1, places=10)

    def test_clone(self, stage="clone"):
        new_nm = self.nm.clone()

        self.assertTrue((self.nm.morphology_data["neuron"].geometry == new_nm.morphology_data["neuron"].geometry).all())
        self.assertTrue((self.nm.morphology_data["neuron"].section_data == new_nm.morphology_data["neuron"].section_data).all())

        # Make sure that clone is a copy, and don't point back to same arrays
        ang = np.pi
        R_x = np.array([[1, 0, 0],
                        [0, np.cos(ang), -np.sin(ang)],
                        [0, np.sin(ang), np.cos(ang)]])

        new_nm.place(rotation=R_x, position=np.array([0, 0, 0]))

        self.assertTrue((np.abs(self.nm.morphology_data["neuron"].geometry[:, 0] - new_nm.morphology_data["neuron"].geometry[:, 0]) < 1e-6).all())
        self.assertTrue((np.abs(self.nm.morphology_data["neuron"].geometry[:, 1] + new_nm.morphology_data["neuron"].geometry[:, 1]) < 1e-6).all())
        self.assertTrue((np.abs(self.nm.morphology_data["neuron"].geometry[:, 2] + new_nm.morphology_data["neuron"].geometry[:, 2]) < 1e-6).all())

        new_nm2 = self.nm.clone()
        new_nm2.place(position=np.array([1, 2, 3]))
        self.assertTrue((np.abs(new_nm2.position - np.array([1, 2, 3])) < 1e-6).all())

    def test_cluster_synapses(self, stage="cluster_synapses"):

        cluster_spread = 30e-6
        n_synapses = 10

        cluster_sec_x, syn_coords, soma_dist = self.nm.cluster_synapses(sec_id=29, sec_x=0.5,
                                                                        count=n_synapses, distance=cluster_spread,
                                                                        rng=np.random.default_rng(20220125))

        self.assertEqual(len(cluster_sec_x), n_synapses)
        self.assertTrue(np.max(soma_dist) - np.min(soma_dist) <= cluster_spread)

        d = scipy.spatial.distance.pdist(syn_coords)
        self.assertTrue(np.all(d <= cluster_spread))
        self.assertTrue(np.any(d >= 10e-6))

        # TODO: Verify self.nm.sec_id_to_len[0] value ... ie segment length of soma?
        # xx = self.nm.compartment_length()
        # np.sum(self.nm.sec_id_to_len[1:])
        # ska vara samma som np.sum(xx)
        # och self.nm.sec_id_to_len[0] deals with the soma.

        if False:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax = self.nm.plot_neuron(plot_axon=False, plot_dendrite=True, axis=ax)
            ax.scatter(syn_coords[:, 0], syn_coords[:, 1], syn_coords[:, 2], s=5, c="red")
            fig.show()
            plt.pause(3)

            import pdb
            pdb.set_trace()


if __name__ == '__main__':
    unittest.main()



import plotly
import plotly.graph_objects as go
import plotly.express as px
import plotly.io as pio
import numpy as np
import os


from snudda.utils import SnuddaLoadSimulation

class PlotReactionDiffusion:

    def __init__(self, network_path, simulation_file, snudda_simulation=None):

        if snudda_simulation is None:
            self.sls = SnuddaLoadSimulation(network_path=network_path,
                                            network_simulation_output_file=simulation_file)
        else:
            self.sls = snudda_simulation

    def list_neuron_info(self, neuron_id):
        return self.sls.list_data_types(neuron_id=neuron_id)

    def plot(self, neuron_id, species=None, species_label=None,
             ylabel=None,
             compartment_id = 0,
             normalise=False,
             fig_name=None, fig_path="figures", title=None,
             width=800, height=700):

        """ compartment_id is based on the order the compartments are added, ie. 0 is first one added, usually soma"""

        if ylabel is None:
            if normalise:
                ylabel = "Concentration (normalised)"
            else:
                ylabel = "Concentration (mM)"

        # pio.renderers.default = "iframe"  # Do not save plots in the notebook, they can get BIG
        # pio.renderers.default = "plotly_mimetype"
        pio.renderers.default = "png"

        palette = px.colors.qualitative.Set2

        fig = go.Figure()

        if species is None:
            species = self.sls.list_data_types(neuron_id=neuron_id, exclude=["spikes", "voltage"])

        time = self.sls.get_time()
        all_data = self.sls.get_all_data(neuron_id=neuron_id, exclude=["spikes", "voltage"])

        for i, s in enumerate(species):
            idx = time >= 0.0

            data = all_data[s]

            color = palette[i % len(palette)]

            if compartment_id is None:
                comp_ofs = 0
            else:
                comp_ofs = np.where(data[1][neuron_id][0] == compartment_id)[0]
                if len(comp_ofs) == 0:
                    print(f"Species {s}, not recorded for section_id {compartment_id}")
                    continue
                else:

                    if len(comp_ofs) > 1:
                        print(
                            f"Warning, species {s} recorded at multiple points in section_id {compartment_id}, using first occurence")

                    comp_ofs = comp_ofs[0]

            if species_label is None:
                s_label = s

                if compartment_id != 0:

                    try:
                        section_id = data[1][neuron_id][0][comp_ofs]
                        section_x = data[1][neuron_id][1][comp_ofs]
                        s_label = f"{s} ({section_id}:{section_x})"

                    except Exception as e:
                        import traceback
                        print(traceback.format_exc())
                        print(e)
                        import pdb
                        pdb.set_trace()

            else:
                s_label = species_label[i]

            try:
                # data variable contains 'data', 'sec_id_x', 'syninfo'
                if normalise:
                    fig.add_trace(go.Scatter(x=time[idx],
                                             y=data[0][neuron_id].T[comp_ofs][idx]/np.max(data[0][neuron_id].T[comp_ofs][idx]),
                                             name=s_label, line={"width": 4, "color": color}))
                else:
                    fig.add_trace(go.Scatter(x=time[idx], y=data[0][neuron_id].T[comp_ofs][idx],
                                             name=s_label, line={"width": 4, "color": color}))
            except Exception as e:
                import traceback
                print(traceback.format_exc())
                print(e)
                import pdb
                pdb.set_trace()

        fig.update_layout(xaxis_title="Time (s)", yaxis_title=ylabel, width=1000, height=800,
                          paper_bgcolor="white", plot_bgcolor="white",
                          font={"size": 15},  # General font size for all elements
                          title={"text": title, "font": {"size": 20}, "x": 0.5, "xanchor": "center", "y": 0.9},
                          legend={"font": {"size": 15}},  # Specific font size for legend
                          xaxis={"title": {"font": {"size": 15}}, "tickfont": {"size": 15}},
                          yaxis={"title": {"font": {"size": 15}},
                                 "tickfont": {"size": 15}}, # Y-axis title and tick labels
                          margin=dict(l=100, r=100, t=160, b=100)) # Margin


        if fig_name is not None:
            if fig_path is not None:
                os.makedirs(fig_path, exist_ok=True)
                fig_name = os.path.join(fig_path, fig_name)

            if ".html" in fig_name:
                fig.write_html(fig_name)
            else:
                fig.write_image(fig_name, width=width, height=height)

        fig.show()

        return fig

def plot_cli():
    import argparse

    parser = argparse.ArgumentParser(description="Plot reaction diffusion data from Snudda simulation")
    parser.add_argument("network_path", type=str, help="Path to the network directory")
    parser.add_argument("simulation_file", type=str, help="Path to the simulation output file")
    parser.add_argument("--neuron_id", type=int, required=True, help="Neuron ID to plot")
    parser.add_argument("--species", type=str, nargs="+", default=None, help="Species to plot")
    parser.add_argument("--species_label", type=str, nargs="+", default=None, help="Labels for species")
    parser.add_argument("--ylabel", type=str, default=None, help="Y-axis label")
    parser.add_argument("--compartment_id", type=int, default=0, help="Compartment ID (default: 0, soma)")
    parser.add_argument("--normalise", action="store_true", help="Normalise the concentration data")
    parser.add_argument("--fig_name", type=str, default=None, help="Output figure filename")
    parser.add_argument("--fig_path", type=str, default="figures",
                        help="Output figure directory (default: figures)")
    parser.add_argument("--title", type=str, default=None, help="Plot title")
    parser.add_argument("--width", type=int, default=800, help="Figure width in pixels (default: 800)")
    parser.add_argument("--height", type=int, default=700, help="Figure height in pixels (default: 700)")

    args = parser.parse_args()

    plotter = PlotReactionDiffusion(
        network_path=args.network_path,
        simulation_file=args.simulation_file
    )

    if args.fig_name is None:
        fig_name = f"plot-{args.neuron_id}.html"
    else:
        fig_name = args.fig_name

    plotter.plot(
        neuron_id=args.neuron_id,
        species=args.species,
        species_label=args.species_label,
        ylabel=args.ylabel,
        compartment_id=args.compartment_id,
        normalise=args.normalise,
        fig_name=fig_name,
        fig_path=args.fig_path,
        title=args.title,
        width=args.width,
        height=args.height
    )

if __name__ == "__main__":
    plot_cli()



import json
import os
import unittest
import numpy as np

from snudda.neurons.neuron_prototype import NeuronPrototype


class NeuronPrototypeTestCase(unittest.TestCase):

    def setUp(self) -> None:

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

    def test_setup(self):

        neuron_path = os.path.join("validation", "striatum-var", "dspn",
                                   "str-dspn-e150917_c9_D1-mWT-1215MSN03-v20211026")

        with self.subTest(msg="neuron_path given"):
            np1 = NeuronPrototype(neuron_path=neuron_path,
                                  neuron_name="np1")

            nm1 = np1.clone(parameter_id=123, morphology_key="m615b0265", modulation_id=789)

            # Check if we load the right neuron
            self.assertEqual(os.path.basename(nm1.swc_filename), "WT-1215MSN03-cor-rep-ax-res3-var0.swc")

            par_file = os.path.join(neuron_path, "parameters.json")
            with open(par_file, "r") as f:
                par_file_data = json.load(f)

            self.assertEqual(np1.get_parameters(parameter_key="p68dce107"), par_file_data["p68dce107"])
            self.assertEqual(np1.get_parameters(parameter_id=7), par_file_data["p30cfaa2c"])  # Parameter ID is modulo

            mod_file = os.path.join(neuron_path, "modulation.json")
            with open(mod_file, "r") as f:
                mod_file_data = json.load(f)

            self.assertEqual(np1.get_modulation_parameters(modulation_key="nm328ca7eb"), mod_file_data["nm328ca7eb"])

            input_info = np1.get_input_parameters(parameter_key="p723b464b", morphology_key="mc257367a")
            self.assertEqual(input_info["cortical"]["nInputs"], 391)
            self.assertEqual(input_info["thalamic"]["nInputs"], 250)

        morphology_path = os.path.join(neuron_path, "morphology")
        parameter_path = os.path.join(neuron_path, "parameters.json")
        modulation_path = os.path.join(neuron_path, "modulation.json")
        mechanism_path = os.path.join(neuron_path, "mechanisms.json")
        meta_path = os.path.join(neuron_path, "meta.json")

        with self.subTest(msg="other paths given"):
            np2 = NeuronPrototype(morphology_path=morphology_path,
                                  parameter_path=parameter_path,
                                  mechanism_path=mechanism_path,
                                  modulation_path=modulation_path,
                                  meta_path=meta_path,
                                  neuron_name="np2", neuron_path=None)

            nm2 = np1.clone(parameter_key="p30cfaa2c", morphology_key="m17af04cc", modulation_key="nmd86e3c46")

            # Check correct morphology was picked
            self.assertEqual(os.path.basename(nm2.swc_filename), "WT-1215MSN03-cor-rep-ax-res3-var6.swc")

# -- compartment_length no longer implemented in NeuronMorphologyExtended (but it was in NeuronMorphology)
#
#        with self.subTest(msg="instantiate all test"):
#            np3 = NeuronPrototype(neuron_path=neuron_path,
#                                  neuron_name="np3")
#
#            np3.instantiate()
#            res = np3.apply("compartment_length", ["dend"])
#
#            # Verify that correct
#            res2 = []
#            for nm in np3.morphology_cache.values():
#                res2.append(nm.compartment_length("dend"))
#
#            res_sum = np.sort([sum(x) for x in res])
#            res2_sum = np.sort([sum(x) for x in res2])
#            self.assertTrue((res_sum == res2_sum).all())
#
#            print(f"Compartment length sums: {res_sum} vs {res2_sum}")


if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
import numpy as np
import json

# Read the parameter data
with open('/media/psf/KTH/2025-11-13-Yvonne-data-teanalysing/Yvonne2019/CategorisedSTP/M1RH-ipsi/M1RH-ipsi_MSN_D1_20Hz_depressing.json-parameters-full.json', 'r') as f:
    data = json.load(f)

# Read the bounds data
with open('/media/psf/KTH/2025-11-13-Yvonne-data-teanalysing/Yvonne2019/CategorisedSTP/M1RH-ipsi/M1RH-ipsi_MSN_D1_20Hz_depressing.json', 'r') as f:
    bounds_data = json.load(f)

bounds = bounds_data['model_data']

# Parameter names
param_names = ['u_sobol', 'tau_r_sobol', 'tau_f_sobol', 'tau_ratio_sobol', 'cond_sobol']
display_names = ['U', 'τ_R', 'τ_F', 'τ_ratio', 'Conductivity']
bound_keys = ['U', 'tauR', 'tauF', 'tauRatio', 'cond']

# Count parameter sets (excluding 'meta' key)
n_sets = len([k for k in data.keys() if k != 'meta'])
set_indices = range(n_sets)

# Organize parameters by type
parameters = {name: [] for name in param_names}
errors = []

for i in range(n_sets):
    params = data[str(i)]['parameters']
    for j, name in enumerate(param_names):
        parameters[name].append(params[j])
    errors.append(data[str(i)]['error'])

# Create figure with subplots
fig, axes = plt.subplots(3, 2, figsize=(14, 12))
fig.suptitle('Parameter Set Analysis - M1RH-ipsi MSN D1 20Hz Depressing', fontsize=16, fontweight='bold', y=0.98)

# Flatten axes for easier indexing
axes_flat = axes.flatten()

# Plot each parameter
for idx, (param_name, display_name, bound_key) in enumerate(zip(param_names, display_names, bound_keys)):
    ax = axes_flat[idx]
    
    # Get values and bounds
    values = parameters[param_name]
    bound_min, bound_max = bounds[bound_key]
    
    # Create bar plot with error-based coloring
    colors = plt.cm.viridis(np.array(errors) / max(errors))
    bars = ax.bar(set_indices, values, color=colors, alpha=0.7, edgecolor='black', linewidth=1.5)
    
    # Add horizontal lines for bounds
    ax.axhline(y=bound_min, color='red', linestyle='--', linewidth=2, alpha=0.5, label=f'Min: {bound_min:.2e}')
    ax.axhline(y=bound_max, color='red', linestyle='--', linewidth=2, alpha=0.5, label=f'Max: {bound_max:.2e}')
    
    # Formatting
    ax.set_xlabel('Parameter Set Index', fontsize=11, fontweight='bold')
    ax.set_ylabel('Value', fontsize=11, fontweight='bold')
    ax.set_title(f'{display_name}', fontsize=13, fontweight='bold')
    ax.set_xticks(set_indices)
    ax.grid(axis='y', alpha=0.3, linestyle='--')
    ax.legend(fontsize=8, loc='best')
    
    # Use scientific notation for conductivity
    if param_name == 'cond_sobol':
        ax.ticklabel_format(axis='y', style='scientific', scilimits=(0,0))

# Plot error values in the last subplot
ax_error = axes_flat[5]
colors_error = plt.cm.viridis(np.array(errors) / max(errors))
bars = ax_error.bar(set_indices, errors, color=colors_error, alpha=0.7, edgecolor='black', linewidth=1.5)
ax_error.set_xlabel('Parameter Set Index', fontsize=11, fontweight='bold')
ax_error.set_ylabel('Error', fontsize=11, fontweight='bold')
ax_error.set_title('Error Values', fontsize=13, fontweight='bold')
ax_error.set_xticks(set_indices)
ax_error.grid(axis='y', alpha=0.3, linestyle='--')

plt.tight_layout(rect=[0, 0.05, 1, 0.96])
plt.subplots_adjust(hspace=0.35, wspace=0.25)

# Add colorbar for error mapping
sm = plt.cm.ScalarMappable(cmap=plt.cm.viridis, norm=plt.Normalize(vmin=min(errors), vmax=max(errors)))
sm.set_array([])
cbar = fig.colorbar(sm, ax=axes_flat, orientation='horizontal', pad=0.08, fraction=0.03, aspect=40)
cbar.set_label('Error Value', fontsize=11, fontweight='bold')
plt.show()

# Print summary statistics
print("Parameter Set Summary:")
print("=" * 60)
for i in range(n_sets):
    print(f"\nSet {i}: Error = {errors[i]:.6f}")
    for param_name, display_name in zip(param_names, display_names):
        print(f"  {display_name:15s}: {parameters[param_name][i]:.6e}")

print("\n" + "=" * 60)
print(f"Best parameter set (lowest error): Set {errors.index(min(errors))} (error = {min(errors):.6f})")
print(f"Worst parameter set (highest error): Set {errors.index(max(errors))} (error = {max(errors):.6f})")

# Print metadata
print("\n" + "=" * 60)
print("Metadata:")
meta = bounds_data['meta_data']
print(f"  Input region: {meta['input_region']}")
print(f"  Cell type: {meta['cell_type']}")
print(f"  Input frequency: {meta['input_frequency']} Hz")
print(f"  Classification: {meta['classification']}")
print(f"  Number of traces: {len(meta['trace_list'])}")



import json
import os


dataPath = "DATA/Planert2010/d1d2conns/traces/"

fileInfo = [ ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-FI.json",
              [4,17,18,21]),
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-ID.json",
              [0,1,10,11,12,13,15,16]),
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-FD.json",
              [2,5,6]),
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-II.json",
              [3,9,19,20,22]),
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-LI.json",
              [7]),
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-DD.json",
              [8]),             
             ("trace_table.txt-parameters.json",
              "trace_table.txt-parameters-DI.json",
              [14])             
]

for fi in fileInfo:
  print("Reading from " + str(fi[0]))
  with open(dataPath + fi[0],"r") as fIn:
    dataIn = json.load(fIn)

    dataOut = dict()

    print("Keeping: ", end="")
    for key in dataIn:
      if int(key) in fi[2]:
        dataOut[key] = dataIn[key]
        print(str(key),end=" ")
    print("")
    
    assert not os.path.exists(dataPath + fi[1]), "File already exists " + dataPath + fi[1]

    print("Writing to " + str(dataPath + fi[1]))
    with open(dataPath + fi[1],"w") as fOut:
      json.dump(dataOut,fOut,indent=2)




import open3d as o3d
import numpy as np


def cut_mesh(path, save=False):
    mesh = o3d.io.read_triangle_mesh(path)
    x, y, z = np.array(mesh.vertices).T
    left = z > np.mean(z)  # apparent mediolateral axis
    left_idxs = np.where(left)[0]
    left_mesh = mesh.select_by_index(left_idxs)
    right_idxs = np.where(~left)[0]
    right_mesh = mesh.select_by_index(right_idxs)

    if save:
        base = path.strip('.obj')
        o3d.io.write_triangle_mesh(filename=base+'-left.obj', mesh=left_mesh)
        o3d.io.write_triangle_mesh(filename=base+'-right.obj', mesh=right_mesh)

    return mesh, right_mesh, left_mesh

if __name__ == '__main__':
    meshes = ['Striatum-v.obj', 'Striatum-d.obj', 'GPe.obj', 'GPi.obj', 'STN.obj', 'SNr.obj', 'SNc.obj']
    for mesh_name in meshes:
        cut_mesh(mesh_name, save=True)



import neuron
import numpy as np

import bluepyopt.ephys as ephys
from snudda.simulate.nrn_simulator_parallel import NrnSimulatorParallel

from snudda.neurons.neuron_model_extended import NeuronModel
from snudda.neurons.neuron_morphology import NeuronMorphology

# Plot all sections
# [neuron.h.psection(x) for x in neuron.h.allsec()]

##############################################################################

class RunSynapseRun(object):
  
  ############################################################################
  
  def __init__(self,
               neuronMorphology,
               neuronMechanisms,
               neuronParameters,
               neuronModulation,
               stimTimes,
               synapseDensity,
               nSynapses,
               synapseSectionID=None, # if given, nSynapses is ignored
               synapseSectionX=None,  # same # of elements as synapseSectionID
               neuronParameterID=0, # Which param set in parameter file to use
               neuronModulationID=0,
               holdingVoltage=-70e-3,
               synapseType='glut',
               params={},
               time=2,
               logFile=None,
               verbose=True,
               rng=None):

    self.logFile = logFile # File pointer
    self.verbose = verbose

    self.writeLog("Holding voltage: " + str(holdingVoltage) + " V")
    self.writeLog("Stim times: " + str(stimTimes) + " s")
    self.writeLog("Synapse type: " + str(synapseType))

    self.time = time
    self.synapses = []
    self.IClamp = None
    self.ncSyn = None

    # Done in NrnSimulatorParallel
    # neuron.h.load_file('stdrun.hoc')

    self.sim = NrnSimulatorParallel(cvode_active=False)
    
    # Should we use weak reference for garbage collection? (weakref package)

    # We load the neuron morphology object also, used to place synapses
    self.writeLog("Using morphology: " + str(neuronMorphology))    
    self.morphology = NeuronMorphology(swc_filename = neuronMorphology)

    # We need to setup the Neuron model
    self.neuron = NeuronModel(param_file=neuronParameters,
                              morph_path=neuronMorphology,
                              mech_file=neuronMechanisms,
                              cell_name="OptimisationNeuron",
                              modulation_file=neuronModulation,
                              parameter_id=neuronParameterID,
                              modulation_id=neuronModulationID)


    self.neuron.instantiate(sim=self.sim)
    self.setRestingVoltage(holdingVoltage*1e3)
    
    neuron.h.celsius = 35
      
    # gnabar_hh: The maximum specific sodium channel conductance [Default value = 0.120 S/cm2]
    # gkbar_hh: The maximum specific potassium channel conductance [Default value = 0.036 S/cm2]
    # gl_hh: The maximum specific leakage conductance [Default value = 0.0003 S/cm2]
    # ena: The reversal potential for the sodium channel [Default value = 50 mV]
    # ek: The reversal potential for the potassium channel [Default value = -77 mV]
    # el_hh: The reversal potential for the leakage channel [Default value = -54.3 mV]

    # We need to set the params also
    self.params = params
    self.defaultCond = 5e-7
    self.rng = rng

    assert rng, " Code has been updated, rng must be a random number generator, eg. rng = np.random.default_rng(random_seed)"

    self.addSynapseDensity(synapseType,synapseDensity,nSynapses,
                           synapseSectionID,synapseSectionX,
                           rng=self.rng)
          
    self.stimTimes = stimTimes*1e3

    # Assumes input in seconds (SI units)
    self.connectInputToSynapses(stimTimes)
    
    self.somaRecord()
    self.synapseCurrentRecord()

    self.updateHoldingCurrent(holdingVoltage)

    #import pdb
    #pdb.set_trace()  
    
  ############################################################################

  def __del__(self):

    # This should not be needed but...
    self.neuron = None
    self.morphology = None
    self.IClamp = None
    self.VClamp = None
    self.vSave  = None
    self.tSave  = None
    self.iSave  = None
    self.ncSyn  = None
    
    self.vecStim    = None
    self.stimVector = None
    self.littleSynapse = None
    
  ############################################################################

  def updateHoldingCurrent(self,holdingVoltage=None):

    self.writeLog("Updating holding current, might take a bit of time")
    
    if(holdingVoltage is None):
      holdingVoltage = self.holdingVoltage
    else:
      self.holdingVoltage = holdingVoltage

    # Disable old iClamp temporarilly
    if(self.IClamp is not None):
      self.IClamp.amp = 0
    
    # Setup a temporary VClamp
    self.VClamp = neuron.h.SEClamp(self.neuron.icell.soma[0](0.5))
    self.VClamp.rs = 1e-9
    self.VClamp.amp1 = holdingVoltage*1e3
    self.VClamp.dur1 = self.time*2*1e3
    # self.writeLog("VClamp duration: " + str(self.VClamp.dur1))

    neuron.h.finitialize(self.holdingVoltage*1e3)
    # !!! There is a WEIRD neuron bug, that if this tstop here is
    # different from duration of simulation, then the *SECOND* time
    # a model is initialised we get the length of tSave set by this
    # value, and not by the tStop of that simulation --- go figure!
    self.setRestingVoltage(self.holdingVoltage*1e3)
    
    neuron.h.tstop = self.time*1e3 # Must set tstop
    neuron.h.run()

    if(False):
      import matplotlib.pyplot as plt
      plt.figure()
      plt.plot(self.t_save, self.v_save)
      plt.title("Holding voltage should be " \
                + str(self.holding_voltage * 1e3) + "mV")
      plt.xlabel("time (ms)")
      plt.ylabel("volt (mV)")
      plt.ion()
      plt.show()
          
      import pdb
      pdb.set_trace()    
    
    cur = float(self.VClamp.i)

    # Remove VClamp
    self.VClamp = None

    if(self.IClamp is None):
      self.IClamp = neuron.h.i_clamp(self.neuron.icell.soma[0](0.5))
    
    # Update current on iClamp
    self.IClamp.amp = cur
    self.IClamp.dur = 2*self.time*1e3

    self.setRestingVoltage(self.holdingVoltage*1e3)

    self.writeLog("Holding voltage " + str(self.holdingVoltage*1e3) + "mV, " \
          + "IClamp amp = " + str(cur) + "nA" )

    
  ############################################################################

  def setStimTimes(self,stimTimes):

    if(len(stimTimes) != len(self.stimTimes) \
       or (stimTimes != self.stimTimes).any()):


      print("Setting stim times to " + str(stimTimes) + " s")
      self.writeLog("Setting stim times to " + str(stimTimes) + " s")
      self.stimVector = neuron.h.Vector(stimTimes*1e3)
      self.stimTimes = stimTimes*1e3


  ############################################################################

  def addSynapseDensity(self,synapseType,
                        synapseDensity,
                        rng,
                        nSynapses=None,
                        plotFlag=False,
                        sectionID=None,
                        sectionX=None):


    if(plotFlag):

      assert sectionID is None and sectionX is None, \
        "Can not plot if sectionID and sectionX are given"
      
    
      inputCoords,sectionID,sectionX,densityFunction,distSynSoma =\
                                    self.morphology.dendrite_input_locations( \
                                                synapse_density=synapseDensity,
                                                num_locations=nSynapses,
                                                return_density=True,
                                                rng=rng)

      self.synapseLocations=inputCoords
      distFromSoma = self.morphology.dend[:,4]

      #plot density function
      plt.figure()
      plt.plot(distFromSoma*1e6,densityFunction,'o')
      plt.xlabel('distance from soma $(\mu m)$')
      plt.title('density function')
      plt.ion()
      plt.show()

      #plot histogram - distance synapses from soma
      plt.figure()
      plt.hist(distSynSoma*1e6,edgecolor='gray', bins=distSynSoma.size)
      plt.xlabel('distance from soma $(\mu m)$')
      plt.ylabel('frequency')
      plt.title('synaptic distance from soma')
      plt.ion()
      plt.show()

    elif(sectionID is None or sectionX is None):

      inputCoords,sectionID,sectionX =\
        self.morphology.dendrite_input_locations(synapse_density=synapseDensity,
                                                 num_locations=nSynapses,
                                                 rng=rng)

      
      self.synapseLocations=inputCoords

    else:
      self.synapseLocations = "Unknown, you specified sectionX and sectionID, "\
                            + "but not synapse xyz coordinates."
      

    # This is so we can find out where the synapses were placed
    self.synapseSectionID = sectionID
    self.synapseSectionX = sectionX
      
    sections = self.neuron.map_id_to_compartment(sectionID)
    
    for s,sx in zip(sections,sectionX):
      self.addSynapse(synapseType,s,sx,self.params)


  ############################################################################
    
  def addSynapse(self,synapseType,section,sectionX,params):

    sectionX = np.maximum(sectionX,1e-6) # Cant be 0 or 1
    
    try:
      if(synapseType.lower() == 'glut'):
        syn = neuron.h.tmGlut(section(sectionX))
      elif(synapseType.lower() == "gaba"):
        syn = neuron.h.tmGabaA(section(sectionX))
      else:
        assert False, "Unknown synapse type: " + str(synapseType)

      self.synapses.append(syn)
        
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)

      self.writeLog("Did you remember to run nrnivmodl first, " \
            + "to generate channels mod files?")
      exit(-1)


    for p in params:

      # Conductance is set separately, it is a property of netcon
      if(p == "cond"):
        self.defaultCond = params["cond"]
        continue

      val = self.SItoNaturalUnits(p,params[p])
        
      setattr(syn,p,val)
      self.writeLog("Setting parameters: " + str(p) + " = " + str(val) \
            + " (neuron natural units)")
      
      
  ############################################################################

  def connectInputToSynapses(self,stimTimes):

    self.writeLog("Stimulation times (s): " + str(stimTimes))
    
    self.ncSyn = []
    
    self.stimVector = neuron.h.Vector(stimTimes*1e3)
    self.vecStim  = neuron.h.VecStim()
    self.vecStim.play(self.stimVector)

    for syn in self.synapses:
      ncs = neuron.h.NetCon(self.vecStim,syn)
      ncs.delay = 0
      ncs.threshold = 0
      self.ncSyn.append(ncs)

  ############################################################################

  def somaRecord(self):

    self.tSave = neuron.h.Vector()
    self.tSave.record(neuron.h._ref_t)

    self.vSave = neuron.h.Vector()
    self.vSave.record(self.neuron.icell.soma[0](0.5)._ref_v)

  ############################################################################

  def synapseCurrentRecord(self):

    self.iSave = []
    
    for syn in self.synapses:
      iRec = neuron.h.Vector()
      iRec.record(syn._ref_i)
      self.iSave.append(iRec)

  ############################################################################
    
  def run(self,tau,tauR,tauF,U,cond=None,time=None):

    assert False, "This is the old run method"
    
    if(time is None):
      time = self.time
    else:
      self.time = time

    if(cond is None):
      cond = self.default_cond
      
    # print(vars())
    
    #print("Running: tau: %.3g, tauR: %.3g, tauF: %.3g, U: %.3g, cond: %.3g\n" \
    #      % (tau,tauR,tauF,U,cond))
    
    
    # Convert from SI units to natural units that Neuron uses
    for ncs in self.nc_syn:
      ncs.weight[0] = 1*cond*1e6

    for syn in self.synapses:
      syn.tau = tau*1e3
      syn.tau_r = tauR * 1e3
      syn.tau_f = tauF * 1e3
      syn.u = U
    
    # print(self.littleSynapse.tau)


    # print("Initialise voltage to " + str(self.holdingVoltage*1e3) + " mV")
    neuron.h.finitialize(self.holding_voltage * 1e3) # OLD : -70
    neuron.h.tstop = time*1e3

    
    neuron.h.run()

    #self.tSave.resize()
    #self.vSave.resize()
    #self.iSave.resize()
    
    if(np.array(self.t_save).shape != np.array(self.v_save).shape):
      self.write_log("THIS IS WRONG, should be same shape!!")
      self.write_log("size t = " + str(np.array(self.t_save).shape))
      self.write_log("size v = " + str(np.array(self.v_save).shape))
      import pdb
      pdb.set_trace()
    
    # print("Last V = " + str(self.vSave[-1]*1e-3))
    
    # Convert back to SI units
    return (np.array(self.t_save) * 1e-3,
            np.array(self.v_save) * 1e-3,
            np.array(self.i_save) * 1e-9)

  ############################################################################

  def setRestingVoltage(self,restVolt):

    self.writeLog("Setting resting voltage to %.3f mV" % restVolt)
    
    soma = [x for x in self.neuron.icell.soma]
    axon = [x for x in self.neuron.icell.axon]
    dend = [x for x in self.neuron.icell.dend]

    cell = soma+axon+dend

    for sec in cell:
      for seg in sec.allseg():
        seg.v = restVolt
  
  ############################################################################

  # I wish Neuron would use natural units...
  
  def SItoNaturalUnits(self,varName,value):

    convFactor = { "U" : 1.0,
                   "tauR" : 1e3,
                   "tauF" : 1e3,
                   "cond" : 1e6,
                   "tau" : 1e3,
                   "nmda_ratio" : 1.0 }

    if varName not in convFactor:
      self.writeLog("Missing conversion fractor for " + str(varName) \
            + ". Please update SItoNaturalUnits function.")
      self.writeLog("convFactor = " + str(convFactor))
      import pdb
      pdb.set_trace()

    try:
      return value*convFactor[varName]
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)
      import pdb
      pdb.set_trace()
    
    
    
  ############################################################################
  def plot(self):

    ax=self.morphology.plot_neuron(axon_colour='red', dend_colour='blue', soma_colour='green')

    ax.scatter(self.synapseLocations[:,0],
               self.synapseLocations[:,1],
               self.synapseLocations[:,2],
               color='black', 
               marker='o',
               s=20)
    plt.ion()    


  ############################################################################
  
  # pars = { "tau1" : 0.25e-3 }
  # The code converts to natural units internally, if your parameter is missing
  # then update SItoNaturalUnits

  # OBS, soma parameters are ignored by run2 (they should be set during setup)
  
  def run2(self,pars,time=None,cond=1e-8):

    self.writeLog("Running with pars: " + str(pars))
    
    if(time is None):
      time = self.time
    else:
      self.time = time

    for p in pars:
      
      if(p == "cond"):
        cond = self.SItoNaturalUnits(p,pars[p])
          
      else:
        v = self.SItoNaturalUnits(p,pars[p])
        for s in self.synapses:
          setattr(s,p,v)
        
    neuron.h.finitialize(self.holdingVoltage*1e3)  
    for ncs in self.ncSyn:
      ncs.weight[0] = cond

    # self.writeLog("Synapse conductance: " + str(cond) + " uS")

    self.setRestingVoltage(self.holdingVoltage*1e3)

#    print("Check voltage")
#    import pdb
#    pdb.set_trace()

    neuron.h.v_hold = self.holdingVoltage * 1e3
    neuron.h.tstop = time*1e3
    neuron.h.run()

    # Convert results back to SI units
    return (np.array(self.tSave)*1e-3,
            np.array(self.vSave)*1e-3,
            np.array(self.iSave)*1e-9)

############################################################################

  def writeLog(self,text,flush=True): # Change flush to False in future, debug
    if(self.logFile is not None):
      self.logFile.write(text + "\n")
      
      if(self.verbose):
        print(text)
        
      if(flush):
        self.logFile.flush()
    else:
      if(self.verbose):
        print(text)  

  
##############################################################################
  
if __name__== "__main__":
  
  import matplotlib.pyplot as plt
  import time
  start = time.time() 
  
  stimTimes = np.array([0.1,0.15,0.2,0.25,0.3,0.35,0.4,0.45,1.0]) + 0.3

  #synapseDensity="0.05/(1+np.exp(-(d-30e-6)/5e-6))"  
  # synapseDensity="0.05/(1+np.exp(-(d-150e-6)/5e-6))"

  synapseDensity="np.exp(-((d-50e-6)/15e-6)**2)"  
  holdingVoltage = -79.5e-3
  
  # !!! OBS changed number of synapses, before was 20 -- cond 1nS
  tau  = 10e-3
  tauR = 200e-3
  tauF = 900e-3
  U    = 0.5

  tau = 0.1
  tauR = 0.1
  tauF = 0.1
  U = 0.1
  cond = 1e-9
  
  rlsr = RunSynapseRun(stimTimes=stimTimes,
                       synapseType="glut",
                       neuronMorphology="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/WT-1215MSN03-cor-rep-ax2.swc",
                       neuronParameters="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/parameters.json",
                       neuronMechanisms="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/mechanisms.json",
                       neuronModulation="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/modulation.json",
                       synapseDensity=synapseDensity,
                       nSynapses=10,
                       neuronParameterID=0,
                       neuronModulationID=0,
                       holdingVoltage=holdingVoltage)

  # I would like to plot the morphology
  # ax = rlsr.plot()

  #import pdb
  #pdb.set_trace()

  params = { "U" : 0.5,
             "tauR" : 0.6,
             "tauF" : 2.0, #1.1
             "tau" : 0.08,
             "cond" : 1e-9,
             "nmda_ratio" : 0.5 }
  

  tS,vS,iS = rlsr.run2(pars=params)
  plt.figure()
  plt.plot(tS,vS)
  plt.xlabel("Time (s)")
  plt.ylabel("Voltage (V)")

  if(True):
    expData = 'DATA/YvonneJohansson2019/M1LH_Analysis_191001.h5'
    expCellID = 131

    import LoadExpData

    led = LoadExpData.LoadExpData(expData)
    cellType = led.get_cell_type(expCellID)
    plt.title("ExpData is cellType : " + str(cellType))
    expV,expT = led.getData("GBZ_CC_H20",expCellID)
    plt.plot(expT,expV,'r-')
  
  if(False):
    plt.figure()
  
    for iiS in iS:
      plt.plot(tS,iiS)
      
    plt.xlabel("Time (s)")
    plt.ylabel("Current (A)")

  plt.ion()
  plt.show()

    
  end = time.time()
  print(end - start)


  import pdb
  pdb.set_trace()



import os
import sys
import unittest

import numpy as np
import scipy

from snudda.detect.detect import SnuddaDetect
from snudda.detect.prune import SnuddaPrune
from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.place.place import SnuddaPlace
from snudda.utils.reposition_neurons import RepositionNeurons
from snudda.utils.load import SnuddaLoad


class TestClusteredPruning(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        self.network_path_unclustered = os.path.join(os.path.dirname(__file__),
                                                     "networks", "network_testing_unclustered_pruning")
        self.network_path_clustered = os.path.join(os.path.dirname(__file__),
                                                   "networks", "network_testing_clustered_pruning")

        self.create_network(self.network_path_clustered)
        self.create_network(self.network_path_unclustered)

    def create_network(self, network_path):

        create_cube_mesh(file_name=os.path.join(network_path, "mesh", "simple_mesh.obj"),
                         centre_point=(0, 0, 0),
                         side_len=500e-6)

        config_file = os.path.join(network_path, "network-config.json")
        position_file = os.path.join(network_path, "network-neuron-positions.hdf5")
        save_file = os.path.join(network_path, "voxels", "network-putative-synapses.hdf5")

        #  TODO: If d_view is None code run sin serial, add test parallel
        sp = SnuddaPlace(config_file=config_file, d_view=None, verbose=True)
        sp.place()

        rpn = RepositionNeurons(position_file=position_file)
        # Post synaptic neurons
        rpn.place(0, position=[0, 0, 0], rotation=np.eye(3))
        rpn.place(1, position=[0, 50e-6, 0], rotation=np.eye(3))
        rpn.place(2, position=[0, 100e-6, 0], rotation=np.eye(3))
        rpn.place(3, position=[0, 150e-6, 0], rotation=np.eye(3))

        ang = np.pi / 2
        R_z = np.array([[np.cos(ang), -np.sin(ang), 0],
                        [np.sin(ang), np.cos(ang), 0],
                        [0, 0, 1]])

        # Pre synaptic neurons
        rpn.place(4, position=[200e-6, 0, 0], rotation=R_z)
        rpn.place(5, position=[200e-6, 50e-6, 0], rotation=R_z)
        rpn.place(6, position=[200e-6, 100e-6, 0], rotation=R_z)
        rpn.place(7, position=[200e-6, 150e-6, 0], rotation=R_z)

        rpn.close()

        # We want to load in the ball and stick neuron that has 20 micrometer soma diameter, and axon (along y-axis),
        # and dendrite along (x-axis) out to 100 micrometer distance from centre of soma.

        sd = SnuddaDetect(network_path=network_path, rc=None,
                          hyper_voxel_size=130, verbose=True)
        sd.detect()

        if False:
            sd.plot_hyper_voxel(plot_neurons=True, fig_file_name="touch-detection-clusters-validation")

        sp = SnuddaPrune(network_path=network_path, verbose=True)
        sp.prune()

    def cluster_compare(self):

        clustered_data_loader = SnuddaLoad(network_file=self.network_path_clustered, load_synapses=True)
        clustered_synapses = clustered_data_loader.data["synapses"]
        clustered_voxel_size = clustered_data_loader.data["voxel_size"]
        clustered_simulation_origo = clustered_data_loader.data["simulation_origo"]

        nonclustered_data_loader = SnuddaLoad(network_file=self.network_path_unclustered, load_synapses=True)
        nonclustered_synapses = nonclustered_data_loader.data["synapses"]
        nonclustered_voxel_size = nonclustered_data_loader.data["voxel_size"]
        nonclustered_simulation_origo = nonclustered_data_loader.data["simulation_origo"]

        for post_id, pre_id in zip([0, 1, 2, 3], [4, 5, 6, 7]):
            clustered_idx = np.where(np.logical_and(clustered_synapses[:, 0] == pre_id, clustered_synapses[:, 1] == post_id))[0]
            nonclustered_idx = np.where(np.logical_and(nonclustered_synapses[:, 0] == pre_id, nonclustered_synapses[:, 1] == post_id))[0]

            clustered_synapse_coords = clustered_synapses[clustered_idx, 2:5] * clustered_voxel_size + clustered_simulation_origo
            nonclustered_synapse_coords = nonclustered_synapses[nonclustered_idx, 2:5] * nonclustered_voxel_size + nonclustered_simulation_origo

            clustered_synapse_dist = scipy.spatial.distance.cdist(clustered_synapse_coords, clustered_synapse_coords)
            nonclustered_synapse_dist = scipy.spatial.distance.cdist(nonclustered_synapse_coords, nonclustered_synapse_coords)

            clustered_n_close = np.sum(clustered_synapse_dist < (10e-6 / clustered_voxel_size))
            nonclustered_n_close = np.sum(nonclustered_synapse_dist < (10e-6 / nonclustered_voxel_size))

            print(f"clustered = {clustered_n_close}, nonclustered: {nonclustered_n_close}")
            self.assertTrue(clustered_n_close > nonclustered_n_close * 1.5)

    def test_clustered(self):

        data_loader = SnuddaLoad(network_file=self.network_path_clustered, load_synapses=True)
        synapses = data_loader.data["synapses"]
        voxel_size = data_loader.data["voxel_size"]
        simulation_origo = data_loader.data["simulation_origo"]

        for post_id, pre_id in zip([0, 1, 2, 3], [4, 5, 6, 7]):
            idx = np.where(np.logical_and(synapses[:, 0] == pre_id, synapses[:, 1] == post_id))[0]
            coords = synapses[idx, 2:5]*voxel_size + simulation_origo

            max_dist = np.linalg.norm(np.max(coords, axis=0) - np.min(coords, axis=0))
            self.assertTrue(max_dist < 50e-6)

    def test_unclustered(self):

        data_loader = SnuddaLoad(network_file=self.network_path_unclustered, load_synapses=True)
        synapses = data_loader.data["synapses"]
        voxel_size = data_loader.data["voxel_size"]
        simulation_origo = data_loader.data["simulation_origo"]

        for post_id, pre_id in zip([0, 1, 2, 3], [4, 5, 6, 7]):
            idx = np.where(np.logical_and(synapses[:, 0] == pre_id, synapses[:, 1] == post_id))[0]
            coords = synapses[idx, 2:5]*voxel_size + simulation_origo

            max_dist = np.linalg.norm(np.max(coords, axis=0) - np.min(coords, axis=0))
            self.assertTrue(max_dist > 120e-6)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from snudda.utils.load import SnuddaLoad
from scipy.spatial import distance_matrix
import matplotlib.pyplot as plt


class PlotDistanceStatistics:

    def __init__(self, network_path):

        self.snudda_load = SnuddaLoad(network_file=network_path)

    def calculate_distance_matrix(self, neuron_type, neuron_type2=None):

        if neuron_type2 is None:
            neuron_type2 = neuron_type

        neuron_id = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type)
        neuron_id2 = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type2)

        positions = self.snudda_load.data["neuronPositions"][neuron_id, :]
        positions2 = self.snudda_load.data["neuronPositions"][neuron_id2, :]

        dist_matrix = distance_matrix(positions, positions2)

        return dist_matrix

    def plot_distance_histogram(self, neuron_type, neuron_type2=None):

        dist_matrix = self.calculate_distance_matrix(neuron_type=neuron_type, neuron_type2=neuron_type2)

        # Nan the diagonal elements
        dist_matrix += np.diag(np.full((dist_matrix.shape[0],), np.nan))

        plt.figure()
        plt.hist(dist_matrix.flatten()*1e6, bins=30)
        plt.xlabel("Distance to neighbour ($\mu$m)")
        plt.ylabel("Count")
        if neuron_type2 is None:
            plt.title(f"Distance between {neuron_type} (mean: {np.nanmean(dist_matrix)*1e6:.1f})")
        else:
            plt.title(f"Distance between {neuron_type} and {neuron_type2} (mean: {np.nanmean(dist_matrix)*1e6:.1f})")

        # plt.ion()
        plt.show()

    def plot_distance_histogram_closest(self, neuron_type, neuron_type2=None, n_closest=5):

        dist_matrix = self.calculate_distance_matrix(neuron_type=neuron_type, neuron_type2=neuron_type2)

        # Nan the diagonal elements
        dist_matrix += np.diag(np.full((dist_matrix.shape[0],), np.nan))
        partial_dist_matrix = np.sort(dist_matrix, axis=1)[:, :n_closest]

        plt.figure()
        plt.hist(partial_dist_matrix.flatten()*1e6, bins=30)
        plt.xlabel("Distance to neighbour ($\mu$m)")
        plt.ylabel("Count")
        if neuron_type2 is None:
            plt.title(f"Distance between {neuron_type} ({n_closest} neighbours) (mean: {np.nanmean(partial_dist_matrix)*1e6:.1f})")
        else:
            plt.title(f"Distance between {neuron_type} and {neuron_type2}  ({n_closest} neighbours) (mean: {np.nanmean(partial_dist_matrix)*1e6:.1f})")

        # plt.ion()
        plt.show()


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser("Plot distance statistics")
    parser.add_argument("network_path", help="Path to neuron folder (or the network-synapse.hdf5 file)")
    parser.add_argument("neuron_type", help="Neuron type to do statistics for", type=str)
    parser.add_argument("--neuron_type2", help="Second neuron type", default=None, type=str)
    parser.add_argument("--closest", help="Only include distance to N closest neighbours", type=int)

    args = parser.parse_args()

    pds = PlotDistanceStatistics(network_path=args.network_path)

    if args.closest is None:
        pds.plot_distance_histogram(neuron_type=args.neuron_type, neuron_type2=args.neuron_type2)
    else:
        pds.plot_distance_histogram_closest(neuron_type=args.neuron_type, neuron_type2=args.neuron_type2,
                                            n_closest=args.closest)





import h5py
import nixio

class NixExport:

    def __init__(self, snudda_file, nix_file: str):

        self.snudda_file = h5py.File(snudda_file, "r")
        self.nix_file = nixio.File.open(nix_file,
                                        nixio.FileMode.ReadWrite,
                                        compression=nixio.Compression.DeflateNormal)
        self.data_block = self.nix_file.blocks("Snudda Simulation")


    def convert(self):

        meta_section = self.nix_file.create_section("simulation_metadata", "nix.simulation_metadata")

        for data_name in self.snudda_file["meta_data"].keys():
            meta_section.create_property(data_name, self.snudda_file["meta_data"][data_name][()])





# This script visualises two neurons and the location of the synapses
# connecting them.

import bpy
import mathutils
import numpy as np
import h5py
from snudda.utils.snudda_path import snudda_parse_path, get_snudda_data
import os

#network_file = "/home/hjorth/HBP/example/network/delme/network-synapses.hdf5"
network_file = "/home/hjorth/HBP/example/network/delme2/network-synapses.hdf5"
network_path = os.path.basename(network_file)

snudda_data = get_snudda_data(network_path=network_path)

out_file = "two-neurons-synapses.blend"

visualise_id = [0, 3]

# Load the neuron positions
fi = h5py.File(network_file, "r")

neuron_id = fi["network/neurons/neuronID"].value
keep_idx = np.array([(x in visualise_id) for x in neuron_id])
neuron_id = neuron_id[keep_idx]

pos = fi["network/neurons/position"][keep_idx, :]
rot = fi["network/neurons/rotation"][keep_idx, :]
morph = fi["network/neurons/morphology"][keep_idx]
name = fi["network/neurons/name"][keep_idx]

origo = fi["meta/simulationOrigo"].value
voxelSize = fi["meta/voxelSize"].value

synapses = fi["network/synapses"]

# Remove the start cube
bpy.ops.object.delete()

bpy.data.scenes['Scene'].render.engine = 'CYCLES'
world = bpy.data.worlds['World']
world.use_nodes = True

# changing these values does affect the render.
bg = world.node_tree.nodes['Background']
if False:  # Set to True for white background
    bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
    bg.inputs[1].default_value = 1.0
else:
    bg.inputs[0].default_value[:3] = (0.0, 0.0, 0.0)
    bg.inputs[1].default_value = 0.0

mat_dspn = bpy.data.materials.new("PKHG")
# matMSD1.diffuse_color = (1.0,0.2,0.2)
mat_dspn.diffuse_color = (77. / 255, 151. / 255, 1.0)

mat_ispn = bpy.data.materials.new("PKHG")
# matMSD2.diffuse_color = (0.2,0.2,1.0)
mat_ispn.diffuse_color = (67. / 255, 55. / 255, 181. / 255)

mat_fs = bpy.data.materials.new("PKHG")
mat_fs.diffuse_color = (6. / 255, 31. / 255, 85. / 255)
mat_chin = bpy.data.materials.new("PKHG")
mat_chin.diffuse_color = (252. / 255, 102. / 255, 0.0)
mat_lts = bpy.data.materials.new("PKHG")
mat_lts.diffuse_color = (150. / 255, 63. / 255, 212. / 255)

mat_other = bpy.data.materials.new("PKHG")
mat_other.diffuse_color = (0.4, 0.4, 0.4)

mat_synapse = bpy.data.materials.new("PKHG")
# matSynapse.diffuse_color = (0.3,1.0,0.3)
mat_synapse.diffuse_color = (1.0, 1.0, 0.9)
# matSynapse.use_transparency = True
mat_synapse.use_nodes = True
# import pdb
# pdb.set_trace()

# Make synapses glow
emission = mat_synapse.node_tree.nodes.new('ShaderNodeEmission')
emission.inputs['Strength'].default_value = 5.0

material_output = mat_synapse.node_tree.nodes.get('Material Output')
mat_synapse.node_tree.links.new(material_output.inputs[0], emission.outputs[0])

for ps, rt, mo, nm, n_id in zip(pos, rot, morph, name, neuron_id):

    e_rot = mathutils.Matrix(rt.reshape(3, 3)).to_euler()
    bpy.ops.import_mesh.swc(filepath=snudda_parse_path(mo, snudda_data))
    obj = bpy.context.selected_objects[0]

    obj.rotation_euler = e_rot
    print(f"Setting position: {ps * 1e3}")
    obj.location = ps * 1e3

    n_type = nm.decode().split("_")[0]
    if n_type == "dSPN" or n_type == "MSD1":
        print(f"{n_id} dSPN")
        mat = mat_dspn
    elif n_type == "iSPN" or n_type == "MSD2":
        print(f"{n_id} iSPN")
        mat = mat_ispn
    elif n_type == "FS":
        print(f"{n_id} FS")
        mat = mat_fs
    elif n_type == "ChIN":
        print(f"{n_id} ChIN")
        mat = mat_chin
    elif n_type == "LTS":
        print(f"{n_id} LTS")
        mat = mat_lts
    else:
        print(f"{n_id} other")
        mat = mat_other

    for ch in obj.children:
        ch.active_material = mat

    obj.select = False

# Draw the synapses

nSynapses = 0

for ob in bpy.context.selected_objects:
    ob.select = False

for syn in synapses:
    pre_id = syn[0]
    post_id = syn[1]

    if pre_id in visualise_id and post_id in visualise_id:
        # Draw this neuron (the SWC import scales from micrometers to mm), the
        # positions in the simulation are in meters, need to scale it to mm for
        # blender to have same units.
        x = (origo[0] + voxelSize * syn[2]) * 1e3
        y = (origo[1] + voxelSize * syn[3]) * 1e3
        z = (origo[2] + voxelSize * syn[4]) * 1e3

        # bpy.ops.mesh.primitive_cube_add(location=(x,y,z))
        # bpy.context.scene.objects.active.dimensions = \
        #  (voxelSize, voxelSize, voxelSize)

        bpy.ops.mesh.primitive_uv_sphere_add(location=(x, y, z), size=0.001)
        ob = bpy.context.selected_objects[0]
        ob.active_material = mat_synapse
        ob.select = False

        nSynapses += 1

        print(f"Added synapse at {[x, y, z]}")

print(f"nSynapses = {nSynapses}")

# import pdb
# pdb.set_trace()

# Add a light source

lamp_data = bpy.data.lamps.new(name="Sun", type='SUN')
lamp_object = bpy.data.objects.new(name="Sun", object_data=lamp_data)
bpy.context.scene.objects.link(lamp_object)

# Place lamp to a specified location
lamp_object.location = (1000.0, 1000.0, 1000.0)

# Reposition camera
cam = bpy.data.objects["Camera"]
cam.location = (2.98, 2.68, 4.96)
cam.rotation_euler = (1.59, 0, -0.26)

if False:
    bpy.ops.render.render()
    bpy.data.images['Render Result'].save_render(filepath="cube-of-neurons-with-neurites.png")

bpy.ops.wm.save_as_mainfile(filepath=out_file)



# snudda_place.py
#
# Johannes Hjorth, Royal Institute of Technology (KTH)
# Human Brain Project 2019
#
# This open source software code was developed in part or in whole in
# the Human Brain Project, funded from the European Union’s Horizon
# 2020 Framework Programme for Research and Innovation under Specific
# Grant Agreements No. 720270 and No. 785907, No 945539
# (Human Brain Project SGA1, SGA2, SGA3).

# TODO: Let place also decide where separate axonal morphologies should be located, save swc morph, and position + rotation
#       Then detect can just load that info directly. Simplifies load_neuron
#       HDF5 filen, ha en matrix med info för axon morph, axon pos, axon rotation, axon parent

# TODO: We need to define axon_density in meta.json file for neurons that need the axon density
#       Then we need to make sure the code reads this axon density if it exists in meta.json file


import json
import os
import sys

import h5py
import numexpr
import numpy as np
import scipy.cluster
from scipy.interpolate import griddata

from snudda.utils.snudda_path import get_snudda_data
from snudda.neurons.neuron_prototype import NeuronPrototype
# from snudda.place.region_mesh_redux import NeuronPlacer, RegionMeshRedux
from snudda.place.region_mesh_vedo import NeuronPlacer, RegionMesh


from snudda.place.rotation import SnuddaRotate
from snudda.utils.snudda_path import snudda_parse_path, snudda_path_exists, snudda_simplify_path

from snudda.neurons.morphology_data import MorphologyData

''' This code places all neurons in space, but does not setup their
    connectivity. That is done by detect.py and prune.py '''


class SnuddaPlace(object):
    """ Places neurons in 3D space. Use detect to add connections, and prune to remove redundant connections. """

    def __init__(self,
                 config_file=None,
                 network_path=None,
                 snudda_data=None,
                 verbose=False,
                 log_file=None,
                 rc=None,
                 d_view=None,
                 h5libver=None,
                 random_seed=None,
                 griddata_interpolation=False,
                 morphologies_stay_inside=True):

        """
        Constructor.

        Args:
            config_file (str) : Path to config file, e.g. network-config.json in network_path
            network_path (str) : Path to network directory
            snudda_data (str): Path to snudda data
            verbose (bool) : Print extra information on screen
            log_file (str) : Log file for place
            rc : ipyparallel remote client
            d_view : ipyparallel direct view object
            h5libver : Version of h5py library
            random_seed (int) : Numpy random seed
            griddata_interpolation (bool) : Should we interpolate density data (5x slower)

        """

        self.rc = rc
        self.d_view = d_view

        if self.rc and not self.d_view:
            self.d_view = self.rc.direct_view(targets='all')

        if not config_file and network_path:
            config_file = os.path.join(network_path, "network-config.json")

        if not network_path and config_file:
            network_path = os.path.dirname(config_file)

        if not log_file and network_path:
            log_dir = os.path.join(network_path, "log")
            os.makedirs(log_dir, exist_ok=True)
            log_file = open(os.path.join(log_dir, "place-neurons.txt"), "w")

        self.network_path = network_path
        self.config_file = config_file
        self.config = None
        self.morphologies_stay_inside = morphologies_stay_inside

        self.axon_config_cache = None

        self.snudda_data = get_snudda_data(snudda_data=snudda_data,
                                           config_file=self.config_file,
                                           network_path=self.network_path)

        self.verbose = verbose
        self.log_file = log_file

        if self.snudda_data is None:
            self.write_log(f"Warning, snudda_data is not set!")

        if self.network_path:
            self.position_file = os.path.join(self.network_path, "network-neuron-positions.hdf5")
        else:
            self.write_log("No network_path given, not setting position_file. Remember to pass it to write_data.")
            self.position_file = None

        if h5libver is None:
            self.h5libver = "latest"
        else:
            self.h5libver = h5libver

        self.write_log(f"Using hdf5 version: {self.h5libver}")

        self.griddata_interpolation = griddata_interpolation

        # List of all neurons
        self.neurons = []
        self.neuron_prototypes = {}
        self.random_seed = random_seed
        self.random_generator = None
        self.rotate_helper = None

        # This defines the neuron units/channels. The dictionary lists all the
        # members of each unit, the neuronChannel gives the individual neurons
        # channel membership
        self.num_population_units = 1
        self.population_unit_placement_method = "random"
        self.population_units = dict([])
        self.population_unit = None

        # These are the dimensions of our space, d_min also needs a "padding"
        # region outside the space where fake neurons are placed. This is to
        # avoid boundary effects, without padding we would get too high density
        # at the edges
        self.volume = dict([])

    def __del__(self):

        if self.rc:
            # Cleanup memory on workers
            from snudda.utils import cleanup
            cleanup(self.rc, "place")

    ############################################################################

    def place(self):

        """ Place neurons in 3D space. """

        self.parse_config()

        if self.morphologies_stay_inside:
            self.avoid_edges_parallel()

        self.write_data()

    ############################################################################

    def write_log(self, text, flush=True, is_error=False, force_print=False):  # Change flush to False in future, debug

        """
        Writes to log file. Use setup_log first. Text is only written to screen if self.verbose=True,
        or is_error = True, or force_print = True.

        test (str) : Text to write
        flush (bool) : Should all writes be flushed to disk directly?
        is_error (bool) : Is this an error, always written.
        force_print (bool) : Force printing, even if self.verbose=False.
        """

        if self.log_file is not None:
            self.log_file.write(f"{text}\n")
            if flush:
                self.log_file.flush()

        if self.verbose or is_error or force_print:
            print(text, flush=True)

    def close_log_file(self):

        if not isinstance(self.log_file, str) and self.log_file:
            self.log_file.close()
            self.log_file = None

    ############################################################################

    def add_neurons(self,
                    swc_path,
                    num_neurons,
                    param_filename=None,
                    mech_filename=None,
                    modulation=None,
                    reaction_diffusion=None,
                    name="Unnamed",
                    hoc=None,
                    volume_id=None,
                    virtual_neuron=False,
                    axon_density=None,
                    parameter_key=None,
                    morphology_key=None,
                    modulation_key=None,
                    config=None,
                    rng=None):

        """
        Add neurons to volume specified.

        Args:
            swc_path (str): Path to morphology directory (or single morphology)
            num_neurons (int): Number of neurons to add
            param_filename (str): Path to parameter file
            mech_filename (str): Path to mechanism file
            modulation (str): Path to neuromodulation file
            reaction_diffusion (str): Path to RxD reaction diffusion file
            name (str): Name of neuron population, e.g. DSPN (which will become DSPN_0, DSPN_1, etc...)
            hoc (str): Path to hoc file (currently disabled)
            volume_id (str): ID of the volume to place neurons in
            virtual_neuron (bool): Real or virtual neuron, the latter can be used to model axons giving input to network
            axon_density (str): Axon density

            parameter_key (str, optional): Parameter Key to use, default=None (Randomise between all available)
            morphology_key (str, optional): Morphology Key to use, default=None (Randomise between all available)
            modulation_key (str, optional): Key for neuromodulation, default=None (Randomise between all available)
            config (dict): dict representation of network-config.json
        """

        assert volume_id is not None, f"You must specify a volume for neuron {name}"
        assert hoc is None, "Currently only support hoc=None, since we can have multiple parameter, morph combos now"

        if rng is None:
            rng = self.random_generator

        neuron_prototype = NeuronPrototype(neuron_name=name,
                                           neuron_path=None,
                                           snudda_data=self.snudda_data,
                                           morphology_path=swc_path,
                                           parameter_path=param_filename,
                                           mechanism_path=mech_filename,
                                           modulation_path=modulation,
                                           reaction_diffusion_path=reaction_diffusion,
                                           load_morphology=False,
                                           virtual_neuron=virtual_neuron,
                                           verbose=self.verbose)

        neuron_type = name.split("_")[0]
        neuron_positions = self.volume[volume_id]["mesh"].place_neurons(num_neurons, neuron_type)

        first_added = True

        neuron_rotations = self.rotate_helper.get_rotations(volume_name=volume_id, neuron_type=neuron_type,
                                                            neuron_positions=neuron_positions,
                                                            rng=rng)

        # Are there any projections from this neuron type in the config file?
        axon_info = self.generate_extra_axon_info(source_neuron=name,
                                                  position=neuron_positions,
                                                  config=config,
                                                  rng=rng)

        for idx, (coords, rotation) in enumerate(zip(neuron_positions, neuron_rotations)):

            # We set loadMorphology = False, to preserve memory
            # Only morphology loaded for nm then, to get axon and dend
            # radius needed for connectivity

            # Pick a random parameter set
            # parameter.json can be a list of lists, this allows you to select the
            # parameter set randomly
            # modulation.json is similarly formatted, pick a parameter set here
            if parameter_key is None:
                parameter_id = rng.integers(1000000)
            else:
                parameter_id = None

            if modulation_key is None:
                modulation_id = rng.integers(1000000)
            else:
                modulation_id = None

            if morphology_key is None:
                morphology_id = rng.integers(1000000)
            else:
                morphology_id = None

            n = neuron_prototype.clone(position=coords,
                                       rotation=rotation,
                                       morphology_id=morphology_id,
                                       parameter_id=parameter_id,
                                       modulation_id=modulation_id,
                                       parameter_key=parameter_key,
                                       morphology_key=morphology_key,
                                       modulation_key=modulation_key)

            n.neuron_id = len(self.neurons)
            n.volume_id = volume_id

            if axon_density:
                n.axon_density_type = axon_density[0]
                n.axon_density = axon_density[1]
                n.max_axon_radius = axon_density[2]
                n.axon_density_bounds_xyz = axon_density[2]

            if axon_info is not None and len(axon_info) > 0:
                for axon_name, axon_position, axon_rotation, axon_swc in axon_info:
                    n.add_morphology(swc_file=axon_swc[idx],
                                     name=axon_name,
                                     position=axon_position[idx, :],
                                     rotation=axon_rotation[idx, :].reshape((3, 3)),
                                     lazy_loading=True)

            self.neurons.append(n)

            # This info is used by workers to speed things up
            if first_added:
                first_added = False
                self.neuron_prototypes[n.name] = n

    ############################################################################

    def parse_config(self, config_file=None, resort_neurons=False):

        """ Pase network config_file"""

        if config_file is None:
            config_file = self.config_file

        if config_file is None:
            self.write_log("No config file specified.", is_error=True)
            raise ValueError("No config file specified.")

        if not os.path.isfile(config_file):
            self.write_log(f"Missing config file {config_file}", is_error=True)
            raise ValueError(f"Missing config file {config_file}")

        self.write_log(f"Parsing place config file {config_file}")

        with open(config_file, "r") as f:
            config = json.load(f)

        if not config:
            self.write_log("Warning, empty network config.")

        if self.random_seed is None:
            if "random_seed" in config and "place" in config["random_seed"]:
                self.random_seed = config["random_seed"]["place"]
                self.write_log(f"Reading random seed from config file: {self.random_seed}")
        else:
            self.write_log(f"Using random seed provided by command line: {self.random_seed}")

        # Setup for rotations
        self.rotate_helper = SnuddaRotate(self.config_file)

        if self.random_seed:
            self.random_generator = np.random.default_rng(self.random_seed + 115)
        else:
            self.random_generator = np.random.default_rng()

        # Setup random seeds for all regions
        ss = np.random.SeedSequence(self.random_seed)
        all_seeds = ss.generate_state(len(config["regions"]))

        for (region_name, region_data), region_seed in zip(config["regions"].items(), all_seeds):
            total_num_neurons = region_data.get("num_neurons")

            volume_data = region_data["volume"]
            self.volume[region_name] = volume_data

            if "random_seed" not in volume_data:
                self.volume[region_name]["random_seed"] = region_seed
            else:
                region_seed = self.volume[region_name]["random_seed"]

            region_rnd = np.random.default_rng(region_seed + 123)

            if snudda_path_exists(volume_data["mesh_file"], self.snudda_data):
                mesh_file = snudda_parse_path(volume_data["mesh_file"], self.snudda_data)
            elif os.path.isfile(os.path.join(self.network_path, volume_data["mesh_file"])):
                mesh_file = os.path.join(self.network_path, volume_data["mesh_file"])
            else:
                raise ValueError(f"Unable to find mesh_file {volume_data['mesh_file']}")

            if "num_putative_points" in volume_data:
                num_putative_points = int(volume_data["num_putative_points"])
            else:
                num_putative_points = None

            self.volume[region_name]["mesh"] \
                = NeuronPlacer(mesh_path=mesh_file,
                               d_min=self.volume[region_name]["d_min"],
                               random_seed=region_seed,
                               n_putative_points=num_putative_points)

            if "density" in self.volume[region_name]:
                for neuron_type in self.volume[region_name]["density"]:
                    density_func = None

                    if "density_function" in self.volume[region_name]["density"][neuron_type]:
                        density_str = self.volume[region_name]["density"][neuron_type]["density_function"]
                        density_func = lambda x, y, z, d_str=density_str: numexpr.evaluate(d_str)

                    if "density_file" in self.volume[region_name]["density"][neuron_type]:
                        density_file = self.volume[region_name]["density"][neuron_type]["density_file"]

                        # We need to load the data from the file
                        from scipy.interpolate import griddata
                        with open(snudda_parse_path(density_file, self.snudda_data), "r") as f:
                            density_data = json.load(f)

                            assert region_name in density_data and neuron_type in density_data[region_name], \
                                f"Volume {region_name} does not contain data for neuron type {neuron_type}"

                            assert "coordinates" in density_data[region_name][neuron_type] \
                                   and "density" in density_data[region_name][neuron_type], \
                                (f"Missing coordinates and/or Density data for "
                                 f"volume {region_name}, neuron type {neuron_type}")

                            # Convert to SI (* 1e-6)
                            coord = np.array(density_data[region_name][neuron_type]["coordinates"]) * 1e-6
                            density = np.array(density_data[region_name][neuron_type]["density"])

                            if self.griddata_interpolation:
                                density_func = lambda x, y, z, c=coord, d=density: \
                                    griddata(points=c, values=d,
                                             xi=np.array([x, y, z]), method="linear",
                                             fill_value=0).transpose()
                            else:
                                density_func = lambda x, y, z, c=coord, d=density: \
                                    griddata(points=c, values=d,
                                             xi=np.array([x, y, z]), method="nearest",
                                             fill_value=0).transpose()

                    self.volume[region_name]["mesh"].define_density(neuron_type, density_func)

            if "neurons" not in region_data:
                self.write_log(f"No neurons specified for volume {region_name}")

            number_of_added_neurons = 0

            for neuron_type, neuron_data in region_data["neurons"].items():

                model_type = neuron_data.get("neuron_type", "neuron")  # default "neuron"
                # rotation_mode currently not used?!

                axon_density = neuron_data.get("axon_density")

                if "num_neurons" in neuron_data:
                    num_neurons = neuron_data["num_neurons"]
                elif "fraction" in neuron_data:
                    if total_num_neurons is None:
                        raise ValueError(f"If fraction is specified, then total_num_neurons for the region {region_name} must be set")

                    if neuron_data["fraction"] < 0 or neuron_data["fraction"] > 1:
                        raise ValueError(f"{neuron_type}: Neuron 'fraction' must be between 0 and 1.")
                    num_neurons = int(neuron_data["fraction"] * total_num_neurons)
                else:
                    raise ValueError(f"You need to specify 'fraction' or 'num_neurons' for {neuron_type}")

                # print(f"{neuron_type = }, {num_neurons = }")

                n_types = len(neuron_data["neuron_path"])

                if isinstance(num_neurons, int):
                    n_neurons = np.full((n_types, ), int(num_neurons/n_types))

                    extra_n = region_rnd.integers(low=0, high=n_types, size=num_neurons-np.sum(n_neurons))

                    for en in extra_n:
                        n_neurons[en] += 1
                elif len(num_neurons) != n_types:
                    raise ValueError(f"num_neurons can be a scalar or a vector, if it is a vector "
                                     f"({num_neurons = }) then its length must equal number of "
                                     f"neuron_path:s given ({n_types})")
                else:
                    n_neurons = num_neurons

                # RxD reaction diffusion config file
                default_reaction_diffusion = neuron_data.get("reaction_diffusion")

                parameter_key_list = SnuddaPlace.replicate_str(neuron_data.get("parameter_key"),
                                                               n_neurons, f"{neuron_type} parameter_key")
                morphology_key_list = SnuddaPlace.replicate_str(neuron_data.get("morphology_key"),
                                                                n_neurons, f"{neuron_type} morphology_key")
                modulation_key_list = SnuddaPlace.replicate_str(neuron_data.get("modulation_key"),
                                                                n_neurons, f"{neuron_type} modulation_key")

                for (neuron_name, neuron_path), num, parameter_key, morphology_key, modulation_key \
                        in zip(neuron_data["neuron_path"].items(), n_neurons,
                               parameter_key_list, morphology_key_list, modulation_key_list):

                    print(f"{neuron_name = }, {num = }, {neuron_path = }")

                    if neuron_name.split("_")[0] != neuron_type and neuron_name != neuron_type:
                        raise ValueError(f"The keys in neuron_path must be {neuron_name}_X where X is usually a number")

                    morph = os.path.join(neuron_path, "morphology")
                    param = os.path.join(neuron_path, "parameters.json")
                    mech = os.path.join(neuron_path, "mechanisms.json")

                    modulation = os.path.join(neuron_path, "modulation.json")
                    if not snudda_path_exists(modulation, snudda_data=self.snudda_data):
                        modulation = None

                    if default_reaction_diffusion is None:
                        reaction_diffusion = os.path.join(neuron_path, "reaction_diffusion.json")
                        if not snudda_path_exists(reaction_diffusion, snudda_data=self.snudda_data):
                            reaction_diffusion = None
                    else:
                        reaction_diffusion = default_reaction_diffusion

                    if model_type == "virtual":
                        param = None
                        mech = None
                        modulation = None
                        hoc = None
                        virtual_neuron = True
                    else:
                        virtual_neuron = False

                    self.add_neurons(name=neuron_name,
                                     swc_path=morph,
                                     param_filename=param,
                                     mech_filename=mech,
                                     modulation=modulation,
                                     reaction_diffusion=reaction_diffusion,
                                     num_neurons=num,
                                     hoc=None,
                                     volume_id=region_name,
                                     virtual_neuron=virtual_neuron,
                                     axon_density=axon_density,
                                     parameter_key=parameter_key,
                                     morphology_key=morphology_key,
                                     modulation_key=modulation_key,
                                     config=config,
                                     rng=region_rnd)

                    number_of_added_neurons += num

            if total_num_neurons is not None and number_of_added_neurons > total_num_neurons*1.01:
                raise ValueError(f"{region_name} should have {total_num_neurons} but {number_of_added_neurons} were added.")

        self.config_file = config_file
        self.config = config

        # We reorder neurons, sorting their IDs after position
        # -- UPDATE: Now we spatial cluster neurons depending on number of workers
        if resort_neurons:
            self.sort_neurons(sort_idx=self.cluster_neurons(rng=region_rnd))

        if False:  # Debug purposes, make sure neuron ranges are ok
            self.plot_ranges()

        self.define_population_units(config)

    @staticmethod
    def get_var_helper(data, key, default_value=None):
        return data[key] if key in data else default_value

    ############################################################################

    def avoid_edges_parallel(self):

        ss = np.random.SeedSequence(self.random_seed + 100)
        neuron_random_seed = ss.generate_state(len(self.neurons))

        bend_neuron_info = []

        for neuron in self.neurons:
            neuron_type = neuron.name.split("_")[0]
            neuron_config = self.config["regions"][neuron.volume_id]["neurons"][neuron_type]

            if "stay_inside_mesh" in neuron_config and neuron_config["stay_inside_mesh"]:
                volume_id = neuron_config["volume_id"]
                mesh_file = self.config["regions"][volume_id]["volume"]["mesh_file"]

                if isinstance(neuron_config["stay_inside_mesh"], dict):
                    if "k_dist" in neuron_config["stay_inside_mesh"]:
                        k_dist = neuron_config["stay_inside_mesh"]["k_dist"]
                    else:
                        k_dist = 30e-6

                    if "n_random" in neuron_config["stay_inside_mesh"]:
                        n_random = neuron_config["stay_inside_mesh"]["n_random"]
                    else:
                        n_random = 5

                    if "max_angle" in neuron_config["stay_inside_mesh"]:
                        max_angle = neuron_config["stay_inside_mesh"]["max_angle"]
                    else:
                        max_angle = 0.1  # radians
                else:
                    k_dist = 30e-6
                    n_random = 5
                    max_angle = 0.1  # radians

                bend_neuron_info.append((neuron.neuron_id, neuron.name, neuron.swc_filename,
                                         neuron.position, neuron.rotation,
                                         neuron_random_seed[neuron.neuron_id],
                                         volume_id, mesh_file, k_dist, n_random, max_angle))

        bend_morph_path = os.path.join(self.network_path, "modified_morphologies")

        if not os.path.isdir(bend_morph_path):
            os.mkdir(bend_morph_path)

        if self.d_view is None:
            # Make sure we use the same random seeds if we run in serial, as would have been used in parallel

            modified_neurons = self.avoid_edges_helper(bend_neuron_info=bend_neuron_info, network_path=self.network_path)

        else:

            # Make random permutation of neurons, to spread out the edge neurons
            unsorted_neuron_id = self.random_generator.permutation(len(bend_neuron_info))
            bend_neuron_info = [bend_neuron_info[idx] for idx in unsorted_neuron_id]

            with self.d_view.sync_imports():
                from snudda.place import SnuddaPlace

            self.d_view.scatter("bend_neuron_info", bend_neuron_info, block=True)
            self.d_view.push({"config_file": self.config_file,
                              "network_path": self.network_path,
                              "snudda_data": self.snudda_data},
                             block=True)

            cmd_str = f"sp = SnuddaPlace(config_file=config_file,network_path=network_path,snudda_data=snudda_data)"
            self.d_view.execute(cmd_str, block=True)

            cmd_str3 = f"modified_neurons = SnuddaPlace.avoid_edges_helper(bend_neuron_info=bend_neuron_info, network_path=network_path)"
            self.d_view.execute(cmd_str3, block=True)

            modified_neurons = self.d_view.gather("modified_neurons", block=True)

        for neuron_id, new_morphology in modified_neurons:
            # Replace the original morphology with the warped morphology, morphology includes rotation
            self.neurons[neuron_id].swc_filename = new_morphology
            self.neurons[neuron_id].rotation = np.eye(3)

    @staticmethod
    def avoid_edges_helper(bend_neuron_info, network_path):

        # TODO: We need name, swc_file, position, rotation
        # This needs to be passed, since self.neurons is not pickleable...

        try:

            from snudda.place.bend_morphologies import BendMorphologies

            bend_morph = dict()
            bend_morph_path = os.path.join(network_path, "modified_morphologies")

            modified_morphologies = []

            for neuron_id, neuron_name, swc_filename, position, rotation, random_seed, volume_id, mesh_file,\
                k_dist, n_random, max_angle in bend_neuron_info:

                if volume_id not in bend_morph:
                    bend_morph[volume_id] = BendMorphologies(region_mesh=mesh_file, rng=None)

                # Returns None if unchanged
                new_morph_name = os.path.join(bend_morph_path, f"{neuron_name}-{neuron_id}.swc")
                new_morphology = bend_morph[volume_id].edge_avoiding_morphology(swc_file=swc_filename,
                                                                                new_file=new_morph_name,
                                                                                original_position=position,
                                                                                original_rotation=rotation,
                                                                                random_seed=random_seed,
                                                                                k_dist=k_dist,
                                                                                n_random=n_random,
                                                                                max_angle=max_angle)

                if new_morphology:
                    modified_morphologies.append((neuron_id, new_morphology))

        except Exception as e:

            # If the code fails in parallel, let's just write a error file with some info

            import traceback
            import uuid

            file_name = f"error_{uuid.uuid4().hex}.txt"
            error_str = (f"Error in avoid_edges_helper:\n {bend_neuron_info =}, {network_path =}"
                         f"Exception:\n {e}\n\nTraceback:\n{traceback.format_exc()}")

            print(error_str)
            print(f"Writing detailed error to {file_name}")

            # Write the error message to the file
            with open(file_name, "w") as file:
                file.write(error_str)

                for var in ["neuron_id", "swc_filename", "neuron_name", "new_morph_name", "random_seed",
                            "position", "rotation", "k_dist", "n_random", "max_angle", "modified_morphologies"]:

                    if var in locals():
                        value = locals()[var]
                        file.write(f"\n {var} = {value}\n")
                    else:
                        file.write(f"\n '{var}' not defined.")

            raise e

        return modified_morphologies

    ############################################################################

    def generate_extra_axon_info(self, source_neuron, position, config, rng):

        axon_info = []

        if self.axon_config_cache is None:

            self.axon_config_cache = dict()

            for region in config["regions"]:
                for neuron_key, neuron_info in config["regions"][region]["neurons"].items():
                    if "axon_config" in neuron_info:
                        axon_config = snudda_parse_path(neuron_info["axon_config"], self.snudda_data)
                        with open(axon_config, "r") as f:
                            self.axon_config_cache[neuron_key] = json.load(f)

        # Each neuron type has a set of axons to choose from
        source_neuron_type = source_neuron.split("_")[0]

        if source_neuron_type in self.axon_config_cache:
            axon_config = self.axon_config_cache[source_neuron_type]

            for axon_name, axon_data in axon_config.items():

                axon_position, axon_rotation, axon_swc \
                    = self.get_projection_axon_location(source_position=position,
                                                        proj_info=axon_data,
                                                        rng=rng)
                axon_info.append([axon_name, axon_position, axon_rotation, axon_swc])

            return axon_info

        else:
            # No extra axons for neuron
            return []

    #########################################################################

    def generate_extra_axon_info_old(self, source_neuron, position, config, rng):

        axon_info = []

        if self.axon_config_cache is None:

            self.axon_config_cache = dict()

            for neuron_key, neuron_info in config["neurons"].items():
                if "axon_config" in neuron_info:
                    axon_config = snudda_parse_path(neuron_info["axon_config"], self.snudda_data)
                    with open(axon_config, "r") as f:
                        self.axon_config_cache[neuron_key] = json.load(f)

        if source_neuron in self.axon_config_cache:
            axon_config = self.axon_config_cache[source_neuron]

            for axon_name, axon_data in axon_config.items():

                axon_position, axon_rotation, axon_swc \
                    = self.get_projection_axon_location(source_position=position,
                                                        proj_info=axon_data,
                                                        rng=rng)
                axon_info.append([axon_name, axon_position, axon_rotation, axon_swc])

            return axon_info

        else:
            # No extra axons for neuron
            return []

    def get_projection_axon_location(self, source_position, proj_info, rng, patch_hull=True):
        if 'projection' not in proj_info:
            raise KeyError("No 'projection' entry in the projection config!")
        
        proj_cfg = proj_info["projection"]
        if "projection_file" in proj_cfg and \
            ("source" in proj_cfg or \
           "destination" in proj_cfg):
           raise NotImplementedError("Projections should specify either a file or a mapping!")

        elif "source" in proj_cfg \
            and "destination" in proj_cfg:
            source = np.array(proj_cfg["source"])*1e-6
            destination = np.array(proj_cfg["destination"])*1e-6

        elif "projection_file" in proj_cfg :
            with open(proj_cfg["projection_file"], 'r') as f:
                proj_file_data = json.load(f)
            source = np.array(proj_file_data["source"])*1e-6
            destination = np.array(proj_file_data["destination"])*1e-6
            
        else: 
            raise NotImplementedError("Unknown projection configuration!")

        # specify the rotations of the termination zones
        rotation_cfg = proj_info.get("rotation", {})
        if "rotation" in rotation_cfg:
            rotation = np.array(rotation_cfg["rotation"])
            rot_position = destination
        
        elif "rotation_file" in rotation_cfg:
            with open(rotation_cfg["rotation_file"], 'r') as f:
                rotation_data = json.load(f)
            rotation = np.array(rotation_data["rotation"])
            rot_position = np.array(rotation_data["position"])*1e-6
        else:
            rotation = None
            rot_position = None

        target_centres = griddata(points=source,
                                  values=destination,
                                  xi=source_position,
                                  method="linear")

        # this checks if there are values outside of the convex hull
        to_patch = np.where(np.isnan(np.sum(target_centres, axis=1)))[0]

        # and patch missing entries
        if patch_hull and len(to_patch)>0:
            self.write_log(f"Patched {len(to_patch)}/{len(target_centres)}")
            target_centres_patched = griddata(points=source,
                                              values=destination,
                                              xi=source_position,
                                              method="nearest")
            target_centres[to_patch] = target_centres_patched[to_patch]

        # which coordinates to use for selecting rotation
        mapping = rotation_cfg.get('mapping', 'target')
        if mapping == "target":
            xi = target_centres
        elif mapping == "source":
            xi = source_position
        else:
            raise NotImplementedError(f"Unknown mapping '{mapping}'!")

        if rotation is not None:
            target_rotation = griddata(points=rot_position,
                                       values=rotation,
                                       xi=xi, method="linear")
            # if the rotation is specified as a field of rotation vectors, 
            # then these need to be converted to matrices.
            if target_rotation[0].shape == (3,):
                rotation_matrices = \
                [SnuddaRotate.rotation_matrix_from_vectors(np.array([0, 0, 1]), rv).flatten()\
                 for rv in target_rotation] 
                target_rotation = np.array(rotation_matrices)
                
        else:
            target_rotation = [None for x in range(source_position.shape[0])]

        num_axons = len(proj_info["morphologies"])
        axon_id = rng.choice(num_axons, source_position.shape[0])
        axon_swc = [proj_info["morphologies"][x] for x in axon_id]

        return target_centres, target_rotation, axon_swc

    ############################################################################

    def all_neuron_positions(self):

        """ Returns all neuron positions as a n x 3 matrix. """

        n_neurons = len(self.neurons)
        pos = np.zeros((n_neurons, 3))

        for i in range(0, n_neurons):
            pos[i, :] = self.neurons[i].position

        return pos

    ############################################################################

    def all_neuron_rotations(self):

        """ Returns all neuron rotations as a n x 3 x 3 matrix. """

        n_neurons = len(self.neurons)
        rot = np.zeros((n_neurons, 3, 3))

        for i in range(0, n_neurons):
            rot[i, :, :] = self.neurons[i].rotation

        return rot

    ############################################################################

    def all_neuron_names(self):

        """ Returns all neuron names as a list. """

        return map(lambda x: x.name, self.neurons)

    ############################################################################

    def gather_extra_axons(self):

        ax_neuron = []
        ax_name = []
        ax_swc = []
        ax_position = []
        ax_rotation = []

        for n_idx, n in enumerate(self.neurons):
            for md_key, md in n.morphology_data.items():
                if md_key != "neuron":
                    ax_neuron.append(n_idx)  # which neuron does axon belong to
                    ax_name.append(md_key)
                    ax_swc.append(md.swc_file)
                    ax_position.append(md.position)
                    ax_rotation.append(md.rotation.reshape((1, 9)))

        if len(ax_neuron) > 0:
            ax_neuron = np.array(ax_neuron)
            ax_position = np.vstack(ax_position)
            ax_rotation = np.vstack(ax_rotation)

        return ax_neuron, ax_name, ax_position, ax_rotation, ax_swc

    ############################################################################

    def write_data(self, file_name=None):

        """ Writes position data to HDF5 file file_name. """

        if not file_name:
            file_name = self.position_file

        assert len(self.neurons) > 0, "No neurons to save!"

        self.write_log(f"Writing data to HDF5 file: {file_name}")

        pos_file = h5py.File(file_name, "w", libver=self.h5libver)

        with open(self.config_file, 'r') as cfg_file:
            config = json.load(cfg_file)

        # Meta data
        save_meta_data = [(self.config_file, "config_file"),
                          (json.dumps(config), "config"),
                          (snudda_parse_path("$DATA", self.snudda_data), "snudda_data")]

        meta_group = pos_file.create_group("meta")

        for data, data_name in save_meta_data:
            meta_group.create_dataset(data_name, data=data)

        network_group = pos_file.create_group("network")

        # Neuron information
        neuron_group = network_group.create_group("neurons")

        # If the name list is longer than 20 chars, increase S20
        name_list = [n.name.encode("ascii", "ignore") for n in self.neurons]
        str_type = 'S' + str(max(1, max([len(x) for x in name_list])))
        neuron_group.create_dataset("name", (len(name_list),), str_type, name_list,
                                    compression="gzip")

        neuron_id_list = np.arange(len(self.neurons))
        neuron_group.create_dataset("neuron_id", (len(neuron_id_list),),
                                    'int', neuron_id_list)

        volume_id_list = [n.volume_id.encode("ascii", "ignore")
                          for n in self.neurons]
        str_type_vid = 'S' + str(max(1, max([len(x) for x in volume_id_list])))

        neuron_group.create_dataset("volume_id",
                                    (len(volume_id_list),), str_type_vid, volume_id_list,
                                    compression="gzip")

        hoc_list = [snudda_simplify_path(n.hoc, self.snudda_data).encode("ascii", "ignore") if hasattr(n, "hoc") else "" for n in self.neurons]
        max_hoc_len = max([len(x) for x in hoc_list])
        max_hoc_len = max(max_hoc_len, 10)  # In case there are none
        neuron_group.create_dataset("hoc", (len(hoc_list),), f"S{max_hoc_len}", hoc_list,
                                    compression="gzip")

        swc_list = [snudda_simplify_path(n.swc_filename, self.snudda_data).encode("ascii", "ignore") for n in self.neurons]
        max_swc_len = max([len(x) for x in swc_list])

        neuron_group.create_dataset("morphology", (len(swc_list),), data=swc_list,
                                    dtype=h5py.special_dtype(vlen=bytes),
                                    compression="gzip")

        neuron_path = [snudda_simplify_path(n.neuron_path, self.snudda_data).encode("ascii", "ignore") for n in self.neurons]
        max_np_len = max([len(x) for x in neuron_path])
        # neuron_group.create_dataset("neuron_path", (len(neuron_path),), f"S{max_np_len}", neuron_path)
        neuron_group.create_dataset("neuron_path", (len(neuron_path),), data=neuron_path,
                                    dtype=h5py.special_dtype(vlen=str))

        virtual_neuron_list = np.array([n.virtual_neuron for n in self.neurons], dtype=bool)
        virtual_neuron = neuron_group.create_dataset("virtual_neuron",
                                                     data=virtual_neuron_list)

        # Create dataset, filled further down
        neuron_position = neuron_group.create_dataset("position",
                                                      (len(self.neurons), 3),
                                                      "float",
                                                      compression="gzip")
        neuron_rotation = neuron_group.create_dataset("rotation",
                                                      (len(self.neurons), 9),
                                                      "float",
                                                      compression="gzip")

        # Write axons to hdf5 file
        ax_neuron, ax_name, ax_position, ax_rotation, ax_swc = self.gather_extra_axons()

        axon_group = neuron_group.create_group("extra_axons")
        axon_group.create_dataset("parent_neuron", data=ax_neuron)
        axon_group.create_dataset("name", (len(ax_name), ), data=ax_name,
                                  dtype=h5py.special_dtype(vlen=bytes), compression="gzip")
        axon_group.create_dataset("position", data=ax_position)
        axon_group.create_dataset("rotation", data=ax_rotation)
        axon_group.create_dataset("morphology", (len(ax_swc), ), data=ax_swc,
                                  dtype=h5py.special_dtype(vlen=bytes), compression="gzip")

        # TODO: Parent tree info, eller motsvarande, måste sparas också!

        pk_list = [n.parameter_key.encode("ascii", "ignore")
                   if n.parameter_key is not None else ""
                   for n in self.neurons]
        pk_str_type = 'S' + str(max(1, max([len(x) for x in pk_list])))

        mk_list = [n.morphology_key.encode("ascii", "ignore")
                   if n.morphology_key is not None else ""
                   for n in self.neurons]
        mk_str_type = 'S' + str(max(1, max([len(x) for x in mk_list])))

        mok_list = [n.modulation_key.encode("ascii", "ignore")
                    if n.modulation_key is not None else ""
                    for n in self.neurons]
        mok_str_type = 'S' + str(max(1, max([len(x) for x in mok_list])))

        rd_list = [n.reaction_diffusion.encode("ascii", "ignore")
                   if n.reaction_diffusion is not None else ""
                   for n in self.neurons]
        rd_str_type = 'S' + str(max(1, max([len(x) for x in rd_list])))

        neuron_param_key = neuron_group.create_dataset("parameter_key",
                                                       (len(self.neurons),),
                                                       pk_str_type,
                                                       compression="gzip")

        neuron_morph_key = neuron_group.create_dataset("morphology_key",
                                                       (len(self.neurons),),
                                                       mk_str_type,
                                                       compression="gzip")

        neuron_modulation_key = neuron_group.create_dataset("modulation_key",
                                                            (len(self.neurons),),
                                                            mok_str_type,
                                                            compression="gzip")

        reaction_diffusion = neuron_group.create_dataset("reaction_diffusion_file",
                                                         (len(self.neurons),),
                                                         rd_str_type)

        neuron_pos_all = np.zeros((len(self.neurons), 3))
        neuron_rot_all = np.zeros((len(self.neurons), 9))

        neuron_param_key_list = []
        neuron_param_key_idx = []
        neuron_morph_key_list = []
        neuron_morph_key_idx = []
        neuron_mod_key_list = []
        neuron_mod_key_idx = []

        reacdiff_list = []
        reacdiff_key = []

        for (i, n) in enumerate(self.neurons):
            neuron_pos_all[i, :] = n.position
            neuron_rot_all[i, :] = n.rotation.reshape(1, 9)

            if n.parameter_key:
                neuron_param_key_list.append(n.parameter_key)
                neuron_param_key_idx.append(i)

            if n.morphology_key:
                neuron_morph_key_list.append(n.morphology_key)
                neuron_morph_key_idx.append(i)

            if n.modulation_key:
                neuron_mod_key_list.append(n.modulation_key)
                neuron_mod_key_idx.append(i)

            if n.reaction_diffusion:
                reacdiff_list.append(n.reaction_diffusion)
                reacdiff_key.append(i)

        neuron_position[:, :] = neuron_pos_all
        neuron_rotation[:, :] = neuron_rot_all

        if len(neuron_param_key_list) > 0:
            neuron_param_key[neuron_param_key_idx] = neuron_param_key_list

        if len(neuron_morph_key_list) > 0:
            neuron_morph_key[neuron_morph_key_idx] = neuron_morph_key_list

        if len(neuron_mod_key_list) > 0:
            neuron_modulation_key[neuron_mod_key_idx] = neuron_mod_key_list

        if len(reacdiff_list) > 0:
            reaction_diffusion[reacdiff_key] = reacdiff_list

        # Store input information
        if self.population_unit is None:
            # If no population units were defined, then set them all to 0 (= no population unit)
            self.population_unit = np.zeros((len(self.neurons),), dtype=int)

        neuron_group.create_dataset("population_unit_id", data=self.population_unit, dtype=int)

        # Variable for axon density "r", "xyz" or "" (No axon density)
        axon_density_type = [n.axon_density_type.encode("ascii", "ignore")
                             if n.axon_density_type is not None
                             else b""
                             for n in self.neurons]

        ad_str_type2 = "S" + str(max(1, max([len(x) if x is not None else 1
                                             for x in axon_density_type])))
        neuron_group.create_dataset("axon_density_type", (len(axon_density_type),),
                                    ad_str_type2, data=axon_density_type,
                                    compression="gzip")

        axon_density = [n.axon_density.encode("ascii", "ignore")
                        if n.axon_density is not None
                        else b""
                        for n in self.neurons]
        ad_str_type = "S" + str(max(1, max([len(x) if x is not None else 1
                                            for x in axon_density])))

        neuron_group.create_dataset("axon_density", (len(axon_density),),
                                    ad_str_type, data=axon_density, compression="gzip")

        axon_density_radius = [n.max_axon_radius
                               if n.max_axon_radius is not None and n.axon_density_type == "r"
                               else np.nan for n in self.neurons]

        neuron_group.create_dataset("axon_density_radius", data=axon_density_radius)

        # We also need to save axon_density_bounds_xyz, and num_axon points for the
        # non-spherical axon density option

        axon_density_bounds_xyz = np.nan * np.zeros((len(self.neurons), 6))

        for ni, n in enumerate(self.neurons):

            if n.axon_density is None:
                # No axon density specified, skip
                continue

            if n.axon_density_type == "xyz":

                try:
                    axon_density_bounds_xyz[ni, :] = np.array(n.axon_density_bounds_xyz)
                except:
                    import traceback
                    tstr = traceback.format_exc()
                    self.write_log(tstr)

                    self.write_log(f"Incorrect density string: {n.axon_density}")
                    sys.exit(-1)

        neuron_group.create_dataset("axon_density_bounds_xyz", data=axon_density_bounds_xyz)

        pos_file.close()

        self.write_log(f"Write done.")

    ############################################################################

    def define_population_units(self, config):

        """ Defines population units.

        """

        method_lookup = {"random": self.random_labeling,
                         "radial_density": self.population_unit_density_labeling,
                         "mesh": self.population_unit_mesh}

        for region_name in self.config["regions"]:
            if "population_units" in self.config["regions"][region_name]:
                for unit_id in self.config["regions"][region_name]["population_units"]["unit_id"]:
                    self.population_units[unit_id] = []

                neuron_id = self.volume_neurons(region_name)
                method_name = self.config["regions"][region_name]["population_units"]["method"]

                assert method_name in method_lookup, \
                    (f"Unknown population placement method {method_name}. "
                     f"Valid options are {', '.join([x for x in method_lookup])}")

                method_lookup[method_name](self.config["regions"][region_name]["population_units"], neuron_id)

    ############################################################################

    def random_labeling(self, population_unit_info, neuron_id):

        """
        Creates random labeling.

        Args:
             neuron_id (list) : All potential neuron ID
             population_unit_info (dict): Dictionary with "method" = "random"
                                          "unit_id" = ID of population unit
                                          "fraction_of_neurons" = Fraction of neurons in this unit
                                          "neuron_types" = List of neuron types that are in this unit
                                          "structure" = Name of the structure the unit is located in

        """

        self.init_population_units()  # This initialises population unit labelling if not already allocated

        unit_id = population_unit_info["unit_id"]
        fraction_of_neurons = population_unit_info["fraction_of_neurons"]
        neuron_types = population_unit_info["neuron_types"]  # list of neuron types that belong to this population unit
        structure_name = population_unit_info["structure"]

        # First we need to generate unit lists (with fractions) for each neuron type
        units_available = dict()
        for uid, fon, neuron_type_list in zip(unit_id, fraction_of_neurons, neuron_types):
            for nt in neuron_type_list:
                if nt not in units_available:
                    units_available[nt] = dict()
                    units_available[nt]["unit"] = []
                    units_available[nt]["fraction"] = []

                units_available[nt]["unit"].append(uid)
                units_available[nt]["fraction"].append(fon)

        all_neuron_types = [n.name.split("_")[0] for n in self.neurons]
        # Next we check that no fraction sum is larger than 1

        for neuron_type in units_available:
            assert np.sum(units_available[neuron_type]["fraction"]) <= 1, \
                (f"Population unit fraction sum for Neuron type {neuron_type} "
                 f"in structure {structure_name} sums to more than 1.")

            assert (np.array(units_available[neuron_type]["fraction"]) >= 0).all(), \
                f"Population unit fractions must be >= 0. Please check {neuron_type} in {structure_name}"

            cum_fraction = np.cumsum(units_available[neuron_type]["fraction"])

            neurons_of_type = [self.neurons[nid].neuron_id
                               for (nid, n_type) in zip(neuron_id, all_neuron_types)
                               if n_type == neuron_type]

            rand_num = self.random_generator.uniform(size=len(neurons_of_type))

            for nid, rn in zip(neurons_of_type, rand_num):
                # If our randum number is smaller than the first fraction, then neuron in first pop unit
                # if random number is between first and second cumulative fraction, then second pop unit
                # If larger than last cum_fraction, then no pop unit was picked (and we set pop unit 0)
                idx = len(cum_fraction) - np.sum(rn <= cum_fraction)

                if idx < len(cum_fraction):
                    uid = units_available[neuron_type]["unit"][idx]
                    self.population_unit[nid] = uid
                    self.population_units[uid].append(nid)
                else:
                    self.population_unit[nid] = 0

    ############################################################################

    def population_unit_density_labeling(self, population_unit_info, neuron_id):

        """
        Creates population units based on radial density functions.

        Args:
            neuron_id (list) : All potential neuron ID
            population_unit_info (dict): "method" must be "radial_density"
                                         "neuron_types" list of neuron types
                                         "centres" of radial probabilityes, one per neuron type
                                         "num_neurons" : Number of neurons in each population unit, only used with radial_density method
                                         "probability_functions" list of probability functions of r (as str)
                                         "unit_id" ID of population unit
        """

        assert population_unit_info["method"] == "radial_density"
        self.init_population_units()  # This initialises population unit labelling if not alraedy allocated

        neuron_types = population_unit_info["neuron_types"]
        centres = np.array(population_unit_info["centres"])
        probability_functions = population_unit_info["probability_functions"]
        unit_id = np.array(population_unit_info["unit_id"])

        if "num_neurons" in population_unit_info and population_unit_info["num_neurons"] is not None:
            num_neurons = population_unit_info["num_neurons"]

            if np.isscalar(num_neurons):
                num_neurons = np.full((len(centres),), num_neurons)
        else:
            num_neurons = [None for x in centres]

        assert len(neuron_types) == len(centres) == len(probability_functions) == len(unit_id) == len(num_neurons)

        # xyz = self.all_neuron_positions()
        unit_probability = np.zeros(centres.shape[0])

        for nid in neuron_id:

            pos = self.neurons[nid].position
            neuron_type = self.neurons[nid].name.split("_")[0]

            for idx, (centre_pos, neuron_type_list, p_func) \
                    in enumerate(zip(centres, neuron_types, probability_functions)):

                if neuron_type in neuron_type_list:
                    d = np.linalg.norm(pos - centre_pos)
                    unit_probability[idx] = numexpr.evaluate(p_func)
                else:
                    unit_probability[idx] = 0  # That unit does not contain this neuron type

            # Next we randomise membership
            rand_num = self.random_generator.uniform(size=len(unit_probability))
            member_flag = rand_num < unit_probability

            # Currently we only allow a neuron to be member of one population unit
            n_flags = np.sum(member_flag)
            if n_flags == 0:
                self.population_unit[nid] = 0
            elif n_flags == 1:
                self.population_unit[nid] = unit_id[np.where(member_flag)[0][0]]
            else:
                # More than one unit, pick the one that had smallest relative randnum
                idx = np.argmax(np.multiply(np.divide(unit_probability, rand_num), member_flag))
                self.population_unit[nid] = unit_id[idx]

        # Also update dictionary with lists of neurons of that unit
        for uid in unit_id:
            # Channel 0 is unassigned, no channel, poor homeless neurons!
            self.population_units[uid] = np.where(self.population_unit == uid)[0]

        # Finally if num_neurons is specified, we need to reduce the number of neurons belonging to that unit
        for u_id, n_neurons in zip(unit_id, num_neurons):
            if n_neurons is not None:

                assert len(self.population_units[u_id]) >= n_neurons, \
                    f"Unable to pick {n_neurons} for population unit {u_id}, only {len(self.population_units[u_id])} ({neuron_types[np.where(unit_id == u_id)[0][0]]}) available."

                if n_neurons < len(self.population_units[u_id]):

                    perm_nid = np.random.permutation(self.population_units[u_id])
                    keep_nid = perm_nid[:n_neurons]
                    self.population_units[u_id] = keep_nid

                    remove_nid = perm_nid[n_neurons:]
                    for rid in remove_nid:
                        self.population_unit[rid] = 0

                    if 0 in self.population_units:
                        self.population_units[0] = np.sort(np.array(list(set(self.population_units[0]).union(remove_nid))))
                    else:
                        self.population_units[0] = remove_nid

    def population_unit_mesh(self, population_unit_info, neuron_id):

        self.init_population_units()  # This initialises population unit labelling if not already allocated

        unit_id = population_unit_info["unit_id"]
        mesh_file = population_unit_info["mesh_file"]
        fraction_of_neurons = population_unit_info["fraction_of_neurons"]
        neuron_types = population_unit_info["neuron_types"]  # list of neuron types that belong to this population unit
        structure_name = population_unit_info["structure"]

        pos = np.vstack([self.neurons[nid].position for nid in neuron_id])
        model_neuron_types = [self.neurons[nid].name.split("_")[0] for nid in neuron_id]

        member_probability = np.zeros(shape=(pos.shape[0], len(mesh_file)))

        for idx, (mf, frac, nts) in enumerate(zip(mesh_file, fraction_of_neurons, neuron_types)):

            # This checks if neurons are of the types that are included in population unit
            has_nt = np.array([n in nts for n in model_neuron_types], dtype=bool)

            rm = RegionMesh(mf, verbose=self.verbose)
            member_probability[:, idx] = np.logical_and(rm.check_inside(pos), has_nt) * frac

        # If the probability sums to more than 1, then normalise it, otherwise keep smaller
        member_probability = np.divide(member_probability, np.maximum(1, np.sum(member_probability, axis =1).reshape(len(member_probability),1)))
        
        # Also, we need to add population unit 0 as an option, since choice needs P_sum = 1
        full_member_probability = np.zeros(shape=(member_probability.shape[0], member_probability.shape[1]+1))
        full_member_probability[:, :-1] = member_probability
        full_member_probability[:, -1] = np.maximum(1 - np.sum(member_probability, axis=1), 0)
        all_unit_id = unit_id + [0]

        # Normalise to 1
        row_sums = np.sum(full_member_probability, axis=1)
        row_sums = row_sums[:, np.newaxis]
        full_member_probability = full_member_probability / row_sums

        for idx, (nid, P) in enumerate(zip(neuron_id, full_member_probability)):

            uid = self.random_generator.choice(all_unit_id, p=P)
            self.population_unit[nid] = uid

            if uid > 0:
                self.population_units[uid].append(nid)

    ############################################################################

    def init_population_units(self):

        """ Initialise population units. If none are given they are all set to 0."""

        if not self.population_unit:
            # If no population units were defined, then set them all to 0 (= no population unit)
            self.population_unit = np.zeros((len(self.neurons),), dtype=int)

    # TODO: In prune we later need to gather all synapses belonging to each neuron, which means opening
    #       the hypervoxel files that contain the worker's neurons' synapses. Therefore it is good to have
    #       only nearby neurons on the same worker. Come up with a better scheme for sorting neurons.
    #

    def plot_ranges(self):

        from matplotlib import pyplot as plt

        n_workers = len(self.d_view) if self.d_view is not None else 1
        range_borders = np.linspace(0, len(self.neurons), n_workers + 1).astype(int)

        colours = np.zeros((len(self.neurons),))
        r_start = 0
        for idx, r_end in enumerate(range_borders[1:]):
            colours[r_start:r_end] = idx + 1
            r_start = r_end

        xyz = self.all_neuron_positions()

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2], c=colours, alpha=0.5)
        plt.show()

    def cluster_neurons(self, n_trials=5, rng=None):

        """
        Cluster neurons, so that nearby neurons are grouped on same worker, to speed up simulations.

        Args:
            n_trials (int) : Number of trials for k-means clustering (default 5)
        """

        if rng is None:
            rng = self.random_generator

        n_workers = len(self.d_view) if self.d_view is not None else 1
        n_clusters = np.maximum(n_workers * 5, 100)
        n_clusters = np.minimum(n_clusters, len(self.neurons))

        if n_workers > 1:
            self.write_log(f"Neurons order is optimised for {n_workers} workers. "
                           f"For reproducibility always use the same number of workers.")

        xyz = self.all_neuron_positions()

        centroids, labels = scipy.cluster.vq.kmeans2(xyz, n_clusters, minit="points", seed=rng)

        n_centroids = centroids.shape[0]
        assert n_centroids == n_clusters
        cluster_member_list = [[] for x in range(n_centroids)]

        # Find the members of each cluster
        for idx, val in enumerate(labels):
            cluster_member_list[val].append(idx)

        num_neurons = xyz.shape[0]
        range_borders = np.linspace(0, num_neurons, n_workers + 1).astype(int)

        global_centroid_order = list(np.argsort(np.sum(centroids, axis=1)))

        neuron_order = -np.ones((num_neurons,), dtype=int)
        neuron_order_ctr = 0
        range_start = 0

        for range_end in range_borders[1:]:

            # While within a range, we want the clusters closest to the current primary cluster
            current_cluster = global_centroid_order[0]
            d = np.linalg.norm(centroids - centroids[current_cluster, :], axis=1)
            local_centroid_order = list(np.argsort(d))

            while range_start < range_end and len(local_centroid_order) > 0:
                while len(cluster_member_list[local_centroid_order[0]]) == 0:
                    del local_centroid_order[0]
                current_cluster = local_centroid_order[0]

                take_n = np.minimum(len(cluster_member_list[current_cluster]), range_end - range_start)
                neuron_order[neuron_order_ctr:neuron_order_ctr + take_n] = cluster_member_list[current_cluster][:take_n]
                neuron_order_ctr += take_n

                del cluster_member_list[current_cluster][:take_n]
                range_start += take_n

            while len(global_centroid_order) > 0 and len(cluster_member_list[global_centroid_order[0]]) == 0:
                del global_centroid_order[0]

            if len(global_centroid_order) == 0:
                break

        # Sometimes the original cluster is bad? Try again...
        if np.count_nonzero(neuron_order < 0) > 0 and n_trials > 1:
            self.write_log(f"Redoing place:neuron_clustering, {np.count_nonzero(neuron_order < 0)} "
                           f"neurons unaccounted for",
                           is_error=True)
            self.write_log(f"incorrect neuron_order={neuron_order} (printed for debugging)")
            neuron_order = self.cluster_neurons(n_trials=n_trials - 1, rng=rng)

        # TODO: This occured once on Tegner, why did it happen?
        assert np.count_nonzero(neuron_order < 0) == 0, \
            "cluster_neurons: Not all neurons accounted for. Please rerun place."

        # Just some check that all is ok
        assert (np.diff(np.sort(neuron_order)) == 1).all(), "cluster_neurons: There are gaps in the sorting, error"

        # TODO: Verify that sort order is ok

        return neuron_order

    def sort_neurons(self, sort_idx=None):

        """ Sorting neurons. If no argument is given they will be sorted along x,y,z axis.

            To use cluster sorting, use:
                sp.sort_neurons(sort_idx=sp.cluster_neurons())

        Args:
            sort_idx (list, optional) : Sort order

            """

        if sort_idx is None:
            # This changes the neuron IDs so the neurons are sorted along x,y or z
            xyz = self.all_neuron_positions()
            sort_idx = np.lexsort(xyz[:, [2, 1, 0]].transpose())  # x, y, z sort order

        self.write_log("Re-sorting the neuron IDs")

        for new_idx, old_idx in enumerate(sort_idx):
            self.neurons[old_idx].neuron_id = new_idx

        self.neurons = [self.neurons[x] for x in sort_idx]

        for idx, n in enumerate(self.neurons):
            assert idx == self.neurons[idx].neuron_id, \
                "Something went wrong with sorting"

    def volume_neurons(self, volume_id):

        return [n.neuron_id for n in self.neurons if n.volume_id == volume_id]

    @staticmethod
    def replicate_str(string, n_replicas, variable_name=None):

        # variable_name is just used to help with error message if incorrect input

        if string is None or isinstance(string, str):
            rep_str = [string for x in range(len(n_replicas))]
        elif len(string) == n_replicas:
            rep_str = string
        else:
            raise ValueError(f"Expected a str or a list of str of length {len(n_replicas)} for {variable_name}")

        return rep_str

    ############################################################################


if __name__ == "__main__":
    assert False, "Please use snudda.py place networks/yournetwork"



# This code is taken from https://github.com/NeuralEnsemble/elephant/blob/master/elephant/spike_train_correlation.py
# and modified to use numba for faster run time.

import numpy as np
from numba import jit


@jit(nopython=True, fastmath=True, cache=True)
def spike_time_tiling_coefficient(spiketrain_a, spiketrain_b, end_time, start_time=0, dt=0.005):
    """
    Calculates the Spike Time Tiling Coefficient (STTC) as described in
    :cite:`correlation-Cutts2014_14288` following their implementation in C.
    The STTC is a pair wise measure of correlation between spike trains.
    It has been proposed as a replacement for the correlation index as it
    presents several advantages (e.g. it's not confounded by firing rate,
    appropriately distinguishes lack of correlation from anti-correlation,
    periods of silence don't add to the correlation and it's sensitive to
    firing patterns).
    The STTC is calculated as follows:
    .. math::
        STTC = 1/2((PA - TB)/(1 - PA*TB) + (PB - TA)/(1 - PB*TA))
    Where `PA` is the proportion of spikes from train 1 that lie within
    `[-dt, +dt]` of any spike of train 2 divided by the total number of spikes
    in train 1, `PB` is the same proportion for the spikes in train 2;
    `TA` is the proportion of total recording time within `[-dt, +dt]` of any
    spike in train 1, TB is the same proportion for train 2.
    For :math:`TA = PB = 1`and for :math:`TB = PA = 1`
    the resulting :math:`0/0` is replaced with :math:`1`,
    since every spike from the train with :math:`T = 1` is within
    `[-dt, +dt]` of a spike of the other train.
    This is a Python implementation compatible with the elephant library of
    the original code by C. Cutts written in C and available `here
    <https://github.com/CCutts/Detecting_pairwise_correlations_in_spike_trains/
    blob/master/spike_time_tiling_coefficient.c>`_:
    """

    N1 = len(spiketrain_a)
    N2 = len(spiketrain_b)

    if N1 == 0 or N2 == 0:
        index = np.nan
    else:
        TA = run_t(spiketrain_a, end_time=end_time, start_time=start_time, dt=dt)
        TB = run_t(spiketrain_b, end_time=end_time, start_time=start_time, dt=dt)
        PA = run_p(spiketrain_a, spiketrain_b, dt=dt)
        PA = PA / N1
        PB = run_p(spiketrain_b, spiketrain_a, dt=dt)
        PB = PB / N2
        # check if the P and T values are 1 to avoid division by zero
        # This only happens for TA = PB = 1 and/or TB = PA = 1,
        # which leads to 0/0 in the calculation of the index.
        # In those cases, every spike in the train with P = 1
        # is within dt of a spike in the other train,
        # so we set the respective (partial) index to 1.
        if PA * TB == 1:
            if PB * TA == 1:
                index = 1.
            else:
                index = 0.5 + 0.5 * (PB - TA) / (1 - PB * TA)
        elif PB * TA == 1:
            index = 0.5 + 0.5 * (PA - TB) / (1 - PA * TB)
        else:
            index = 0.5 * (PA - TB) / (1 - PA * TB) + 0.5 * (PB - TA) / (1 - PB * TA)

    return index


@jit(nopython=True, fastmath=True, cache=True)
def run_p(spiketrain_a, spiketrain_b, dt=0.005):
    """
    Check every spike in train 1 to see if there's a spike in train 2
    within dt
    """
    N2 = len(spiketrain_b)

    # Search spikes of spiketrain_i in spiketrain_j
    # ind will contain index of
    ind = np.searchsorted(spiketrain_b, spiketrain_a)

    # To prevent IndexErrors
    # If a spike of spiketrain_i is after the last spike of spiketrain_j,
    # the index is N2, however spiketrain_j[N2] raises an IndexError.
    # By shifting this index, the spike of spiketrain_i will be compared
    # to the last 2 spikes of spiketrain_j (negligible overhead).
    # Note: Not necessary for index 0 that will be shifted to -1,
    # because spiketrain_j[-1] is valid (additional negligible comparison)
    ind[ind == N2] = N2 - 1

    # Compare to nearest spike in spiketrain_j BEFORE spike in spiketrain_i
    close_left = np.abs(spiketrain_b[ind - 1] - spiketrain_a) <= dt
    # Compare to nearest spike in spiketrain_j AFTER (or simultaneous)
    # spike in spiketrain_j
    close_right = np.abs(spiketrain_b[ind] - spiketrain_a) <= dt

    # spiketrain_j spikes that are in [-dt, dt] range of spiketrain_i
    # spikes are counted only ONCE (as per original implementation)
    close = close_left + close_right

    # Count how many spikes in spiketrain_i have a "partner" in
    # spiketrain_j
    return np.count_nonzero(close)


@jit(nopython=True, fastmath=True, cache=True)
def run_t(spiketrain, end_time, start_time=0, dt=0.005):
    """
    Calculate the proportion of the total recording time 'tiled' by spikes.
    """
    N = len(spiketrain)
    time_A = 2 * N * dt  # maximum possible time

    if N == 1:  # for only a single spike in the train

        # Check difference between start of recording and single spike
        if spiketrain[0] - start_time < dt:
            time_A += - dt + spiketrain[0] - start_time

        # Check difference between single spike and end of recording
        elif spiketrain[0] + dt > end_time:
            time_A += - dt - spiketrain[0] + end_time

    else:  # if more than a single spike in the train

        # Calculate difference between consecutive spikes
        diff = np.diff(spiketrain)

        # Find spikes whose tiles overlap
        idx = np.where(diff < 2 * dt)[0]
        # Subtract overlapping "2*dt" tiles and add differences instead
        time_A += - 2 * dt * len(idx) + diff[idx].sum()

        # Check if spikes are within +/-dt of the start and/or end
        # if so, subtract overlap of first and/or last spike
        if spiketrain[0] - start_time < dt:
            time_A += spiketrain[0] - start_time - dt
        if (end_time - spiketrain[N - 1]) < dt:
            time_A += - spiketrain[-1] - dt + end_time

    # Calculate the proportion of total recorded time to "tiled" time
    T = time_A / end_time
    return T


if __name__ == "__main__":

    time_a = np.sort(np.random.uniform(low=0, high=10, size=10))
    time_b = np.sort(np.random.uniform(low=0, high=10, size=10))

    run_p(time_a, time_b, dt=0.005)
    run_t(time_a, end_time=10, dt=0.005)

    spike_time_tiling_coefficient(time_a, time_b, end_time=10)




import json

from snudda.utils import SnuddaLoadSimulation

# TODO: Make sure unit conversion is correct.


class ExtractSteadyStateValues:

    def __init__(self, reaction_diffusion_cfg: str, network_simulation,
                 neuron_id=0, output_cfg=None):

        self.simulation_data = SnuddaLoadSimulation(network_simulation_output_file=network_simulation)

        self.reaction_diffusion_cfg = reaction_diffusion_cfg
        self.neuron_id = neuron_id

        if output_cfg is None or output_cfg == "":
            output_cfg = f"{reaction_diffusion_cfg}-updated"

        self.output_cfg = output_cfg

        with open(self.reaction_diffusion_cfg, "rt") as f:
            self.reaction_diffusion_data = json.load(f)

        self.updated_simulation_config_data = self.reaction_diffusion_data.copy()

    def extract(self, neuron_id=None):

        if neuron_id is None:
            neuron_id = self.neuron_id

        if "species" not in self.updated_simulation_config_data:
            raise ValueError(f"No 'species' defined in JSON file")

        list_species = self.simulation_data.list_data_types(neuron_id=neuron_id)

        for species_name, species_data in self.updated_simulation_config_data["species"].items():

            print(f"Parsing species {species_name}")

            for item_name in ["initial_concentration", "concentration"]:
                if item_name not in species_data:
                    raise KeyError(f"Missing {item_name} for species {species_name}")

            if species_data["concentration"] != species_data["initial_concentration"]:
                raise ValueError(f"{species_name} has concentration = {species_data['concentration']}, "
                                 f"initial_concentration = {species_data['initial_concentration']}")

            if species_name not in list_species:
                raise KeyError(f"Species {species_name} missing in RxD data for neuron {neuron_id}")

            sim_data = self.simulation_data.get_data(data_type=species_name, neuron_id=neuron_id)

            # Check that we read from soma
            if sim_data[1][0][0][0] != -1:
                raise ValueError(f"First data column for {species_name} is not soma!")

            steady_state_conc = sim_data[0][0][-1, 0]

            species_data["concentration"] = steady_state_conc
            species_data["initial_concentration"] = steady_state_conc

        print("Parsing done.")

    def write(self):

        print(f"Writing updated config file to {self.output_cfg}")

        with open(self.output_cfg, "wt") as f:
            json.dump(self.updated_simulation_config_data, f, indent=4)


def cli():

    from argparse import ArgumentParser
    parser = ArgumentParser(description="Extract steady state from simulation, and update concentrations "
                                        "in RxD configuration file.")
    parser.add_argument("reaction_diffusion_cfg", type=str)
    parser.add_argument("network_simulation_file", type=str)
    parser.add_argument("--output_cfg", type=str, default="")
    parser.add_argument("--neuron_id", type=int, default=0)
    args = parser.parse_args()

    print("Args parsed")
    essv = ExtractSteadyStateValues(reaction_diffusion_cfg=args.reaction_diffusion_cfg,
                                    network_simulation=args.network_simulation_file,
                                    neuron_id=args.neuron_id,
                                    output_cfg=args.output_cfg)

    essv.extract()
    essv.write()


if __name__ == "__main__":

    cli()


import os
import numpy as np

from snudda import SnuddaLoad
import matplotlib.pyplot as plt


# TODO: We now have kept, removed and added synapses.
#       Plot the neuron morphology of the original and the degenerated neuron
#       plot synapses depending on type

class PlotDegenerationAndGrowth:

    def __init__(self, original_network_path, degenerated_network_path, neuron_id):

        self.neuron_id = neuron_id
        self.original_file = SnuddaLoad(original_network_path)
        self.degenerated_file = SnuddaLoad(degenerated_network_path)

        self.check_same_network()

        self.fig_path = os.path.join(degenerated_network_path, "figures")
        if not os.path.exists(self.fig_path):
            os.mkdir(self.fig_path)

    def plot_synapses(self):

        added_synapse_coords, removed_synapse_coords, kept_synapse_coords = self.categorising_synapses()

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        if removed_synapse_coords.size > 0:
            ax.scatter(removed_synapse_coords[:, 0],
                       removed_synapse_coords[:, 1],
                       removed_synapse_coords[:, 2],
                       marker='.', color="grey")

        if kept_synapse_coords.size > 0:
            ax.scatter(kept_synapse_coords[:, 0],
                       kept_synapse_coords[:, 1],
                       kept_synapse_coords[:, 2],
                       marker=".", color="black", s=50)

        if added_synapse_coords.size > 0:
            ax.scatter(added_synapse_coords[:, 0],
                       added_synapse_coords[:, 1],
                       added_synapse_coords[:, 2],
                       marker=".", color="red")

        plt.ion()
        plt.show()

        # import pdb
        # pdb.set_trace()

        return fig, ax

    def check_same_network(self):

        assert len(self.original_file.data["neurons"]) == len(self.degenerated_file.data["neurons"]), \
            f"NOT THE SAME NETWORK!! Different number of neurons in the two networks, aborting."

        for orig_neuron, degen_neuron in zip(self.original_file.data["neurons"], self.degenerated_file.data["neurons"]):
            for var in ["name", "position", "rotation"]:
                comp = orig_neuron[var] == degen_neuron[var]
                assert (type(comp) == bool and orig_neuron[var] == degen_neuron[var]) \
                    or (not type(comp) == bool and (orig_neuron[var] == degen_neuron[var]).all()), \
                    f"NOT SAME NETWORK!! Mismatch in {var} for {orig_neuron} and {degen_neuron}"

    def categorising_synapses(self):

        orig_pre_synapses, orig_pre_coords = self.original_file.find_synapses(pre_id=self.neuron_id)
        orig_post_synapses, orig_post_coords = self.original_file.find_synapses(post_id=self.neuron_id)

        degen_pre_synapses, degen_pre_coords = self.degenerated_file.find_synapses(pre_id=self.neuron_id)
        degen_post_synapses, degen_post_coords = self.degenerated_file.find_synapses(post_id=self.neuron_id)

        added_coords, removed_coords, kept_coords = self.get_synapse_locations(orig_post_coords, degen_post_coords)

        return added_coords, removed_coords, kept_coords


#        removed_synapses = np.stack(removed_synapses_list)
#        added_synapses = np.stack(added_synapses_list)
#        kept_synapses_A = np.stack([x for x, y in kept_synapses_list])
#        kept_synapses_B = np.stack([y for x, y in kept_synapses_list])
#
#        removed_synapse_coords = removed_synapses[:, 2:5] * self.original_file.data["voxelSize"] + self.original_file.data["simulationOrigo"]
#        kept_synapse_coords_A = kept_synapses_A[:, 2:5] * self.original_file.data["voxelSize"] + self.original_file.data["simulationOrigo"]
#        kept_synapse_coords_B = kept_synapses_B[:, 2:5] * self.degenerated_file.data["voxelSize"] + self.degenerated_file.data["simulationOrigo"]
#        added_synapse_coords = removed_synapses[:, 2:5] * self.degenerated_file.data["voxelSize"] + self.degenerated_file.data["simulationOrigo"]
#
#        return removed_synapses, added_synapses, kept_synapses_A, kept_synapses_B, \
#                removed_synapse_coords, added_synapse_coords, kept_synapse_coords_A, kept_synapse_coords_B


    def get_synapse_locations(self, coords_1, coords_2):

        if coords_1 is None:
            coords_1 = np.zeros((3, 0))

        if coords_2 is None:
            coords_2 = np.zeros((3, 0))

        voxel_size = self.original_file.data["voxel_size"]

        rcoords_1 = np.round(coords_1/voxel_size, decimals=0).astype(int)
        rcoords_2 = np.round(coords_2/voxel_size, decimals=0).astype(int)

        coord_set_1 = set([tuple(x) for x in rcoords_1])
        coord_set_2 = set([tuple(x) for x in rcoords_2])

        removed = coord_set_1.difference(coord_set_2)
        added = coord_set_2.difference(coord_set_1)
        kept = coord_set_2.intersection(coord_set_1)

        print(f"Added: {len(added)}, removed: {len(removed)}, kept: {len(kept)}")

        added_coords = np.array(list(added), dtype=float)*voxel_size
        removed_coords = np.array(list(removed), dtype=float)*voxel_size
        kept_coords = np.array(list(kept), dtype=float)*voxel_size

        # import pdb
        # pdb.set_trace()

        return added_coords, removed_coords, kept_coords

    def categorise_helper(self, orig_synapses, orig_coords, degen_synapses, degen_coords):

        done = False

        removed_synapses = []
        added_synapses = []
        kept_synapses = []

        orig_iter = self.iter_synapses(orig_synapses, orig_coords)
        degen_iter = self.iter_synapses(degen_synapses, degen_coords)

        orig_syn, orig_prio = next(orig_iter, (None, None))
        degen_syn, degen_prio = next(degen_iter, (None, None))

        while not done:
            if orig_prio == degen_prio:
                kept_synapses.append((orig_syn, degen_syn))
                orig_syn, orig_prio = next(orig_iter, (None, None))
                degen_syn, degen_prio = next(degen_iter, (None, None))

            elif degen_syn is None or orig_prio < degen_prio:
                removed_synapses.append(orig_syn)
                orig_syn, orig_prio = next(orig_iter, (None, None))

            elif orig_syn is None or degen_prio < orig_prio:
                added_synapses.append(degen_syn)
                degen_syn, degen_prio = next(degen_iter, (None, None))

            if orig_syn is None and degen_syn is None:
                done = True

        return removed_synapses, added_synapses, kept_synapses

    def iter_synapses(self, synapses, coords):
        for syn_row in synapses:
            yield syn_row, self.calculate_order(syn_row)

    def calculate_order(self, synapse_row):
        if synapse_row is None:
            return None
        else:
            return (synapse_row[1] * self.num_neurons + synapse_row[0])*max(100, self.num_neurons) + synapse_row[6]




import multiprocessing
import os
import unittest
import json
import numpy as np
from snudda.utils.load import SnuddaLoad
from snudda.utils.load_network_simulation import SnuddaLoadSimulation

import traceback
import io
from contextlib import redirect_stdout, redirect_stderr

try:
    multiprocessing.set_start_method("spawn", force=True)
except RuntimeError:
    # Already set, fine
    pass

def run_pair_recording(network_path, log_file=None):
    from neuron import h
    from snudda.simulate.pair_recording import PairRecording
    experiment_config_file = os.path.join(network_path, "experiment.json")
    network_file = os.path.join(network_path, "network-synapses.hdf5")

    pr = PairRecording(network_path=network_path,
                       network_file=network_file,
                       log_file=log_file,
                       experiment_config_file=experiment_config_file)

    # Explicitly load mechanisms compiled by nrnivmodl
    # pr.load_mechanisms()

    pr.run()
    if pr.log_file is not None:
        pr.log_file.close()

def run_pair_recording_safe(network_path, q):

    log_file = "pair_recording_test_log.txt"

    f = io.StringIO()
    try:
        with redirect_stdout(f), redirect_stderr(f):
            run_pair_recording(network_path=network_path, log_file=log_file)
        q.put(("ok", f.getvalue()))
    except Exception:
        q.put(("error", f.getvalue() + "\n" + traceback.format_exc()))
    finally:
        f.close()

    if os.path.exists(log_file):
        print("\n--- PairRecording log file contents ---")
        with open(log_file) as f:
            print(f.read())
        print("--- End of log ---\n")
    else:
        print(f"Log file not found: {log_file}")

class PairRecordingTestCase(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        # Temporarily disable the creation of network while testing...
        # print("test_pair_recoring.py -- Network creation disabled -- we need to retune or verify the if_info.json files before running")
        # return

        self.network_path = os.path.join("networks", "pair_recording_test")
        rc = None

        if not os.path.isdir("x86_64") and not os.path.isdir("aarch64"):
            print(f"Mech files not compiled --- for some reason {os.getcwd()}, compiling!")
            mech_path = os.path.join(os.path.dirname(__file__), "validation/mechanisms")
            os.system(f"nrnivmodl {mech_path}")

        from snudda import SnuddaPlace
        from snudda import SnuddaDetect
        from snudda import SnuddaPrune

        sp = SnuddaPlace(network_path=self.network_path, rc=rc)
        sp.place()
        sp.close_log_file()

        sd = SnuddaDetect(network_path=self.network_path, rc=rc)
        sd.detect()
        sd.close_log_file()

        sp = SnuddaPrune(network_path=self.network_path, rc=rc)
        sp.prune()
        sp.close_log_file()

        self.experiment_config_file = os.path.join(self.network_path, "experiment.json")
        network_file = os.path.join(self.network_path, "network-synapses.hdf5")

        q = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=run_pair_recording_safe,
            args=(self.network_path, q)
        )
        process.start()
        process.join(timeout=300)

        if process.is_alive():
            process.terminate()
            raise RuntimeError("PairRecording timed out after 120s")

        if not q.empty():
            status, output = q.get()
            print("Child process output:\n", output)
            if status == "error":
                raise RuntimeError("PairRecording failed:\n" + output)
        else:
            raise RuntimeError("PairRecording failed: no output received from process")

        # self.pr = PairRecording(network_path=self.network_path, network_file=network_file,
        #                         experiment_config_file=self.experiment_config_file)
        # self.pr.run()


    def test_frequency(self):

        print("test_pair_recording.py -- This UNIT test is not currently active.")

        # TODO: TEMP line, this is normally done in setUp... remove later
        self.network_path = os.path.join("networks", "pair_recording_test")
        self.experiment_config_file = os.path.join(self.network_path, "experiment.json")

        sim_file = os.path.join(self.network_path, "simulation", "pair-recording-simulation.hdf5")

        sl = SnuddaLoad(self.network_path)
        sns = SnuddaLoadSimulation(network_path=self.network_path,
                                   network_simulation_output_file=sim_file)

        with open(self.experiment_config_file, "r") as f:
            experiment_config = json.load(f)

        for inj_info in experiment_config["current_injection"]:

            neuron_id = inj_info["neuron_id"]
            start = inj_info["start"]
            end = inj_info["end"]
            requested_freq = inj_info["requested_frequency"]

            with self.subTest(f"Testing neuron {neuron_id}"):

                spikes = sns.get_spikes(neuron_id)
                for s, e in zip(start, end):
                    n_spikes = len(np.where(np.logical_and(s <= spikes, spikes <= e))[0])
                    freq = n_spikes / (e - s)

                    print(f"neuron_id: {neuron_id}, requested_freq: {requested_freq} Hz, actual freq: {freq}")
                    self.assertTrue(requested_freq * 0.7 < freq < requested_freq * 1.3,
                                    f"neuron_id: {neuron_id}, requested_freq: {requested_freq} Hz, actual freq: {freq}")

#        import pdb
#        pdb.set_trace()

        # TODO: Read in the frequency of each neuron. Compare the simulated frequency to the requested frequency

        self.assertEqual(True, True)  # add assertion here

    # def tearDown(self):
    #     self.pr.clear_neuron()


if __name__ == '__main__':
    unittest.main()



import bpy
import numpy as np

poscol = np.genfromtxt('/home/hjorth/HBP/StriatumNetwork/model/article/hypervoxelcoords.txt', delimiter=',')
# poscol = np.genfromtxt('hypervoxelcoords2.txt', delimiter=',')


# Remove the start cube
bpy.ops.object.delete()


bpy.data.scenes['Scene'].render.engine = 'CYCLES'
world = bpy.data.worlds['World']
world.use_nodes = True

# changing these values does affect the render.
bg = world.node_tree.nodes['Background']
bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
bg.inputs[1].default_value = 1.0

mats = dict([])

for row in poscol:
  x = row[0]*1e4
  y = row[1]*1e4
  z = row[2]*1e4
  col = row[3:6]
  tcol = tuple(col)

  if(tcol in mats):
    m = mats[tcol]
  else:
    m = bpy.data.materials.new("PKHG")  
    m.diffuse_color = col
    mats[tcol] = m

  bpy.ops.mesh.primitive_uv_sphere_add(location=(x,y,z),size=0.01*10)
  o = bpy.context.selected_objects[0] 
  o.active_material = m

# Add a light source

lampData = bpy.data.lamps.new(name="Sun", type='SUN')
lampObject = bpy.data.objects.new(name="Sun", object_data=lampData)
bpy.context.scene.objects.link(lampObject)


# Place lamp to a specified location
lampObject.location = (5.0, 5.0, 15.0)

bpy.context.scene.camera.data.clip_end = 10000


# Reposition camera
cam = bpy.data.objects["Camera"]
#cam.location = (55.965,24.934,36.768)
cam.location = (8.89, 5.23, 3.51 )
cam.rotation_euler = (71.2*np.pi/180,0,-239*np.pi/180)
  



import os
import sys
import numpy as np
import h5py
import pandas
import timeit


class FilterFlagserData:

    def __init__(self, flagser_data_file_name, network_meta_file_name, filtered_file_name=None):

        self.data_file_name = flagser_data_file_name
        self.network_meta_file_name = network_meta_file_name

        self.data_file = None
        self.meta_data = None
        self.filtered_data_file = None

        if filtered_file_name:
            self.filtered_file_name = filtered_file_name
        else:
            self.filtered_file_name = f"{flagser_data_file_name}-filtered.hdf5"

    def close(self):
        if self.data_file:
            self.data_file.close()
            self.data_file = None

    def load_data(self):

        self.data_file = h5py.File(self.data_file_name, "r")
        # self.meta_data = np.genfromtxt(self.network_meta_file_name, delimiter=",")
        self.meta_data = pandas.read_csv(self.network_meta_file_name,
                                         names=["neuron_id", "neuron_type", "neuron_name", "x", "y", "z", "morphology"])

    def find_core_neurons(self, population_fraction=None, distance_to_centre=None):

        neuron_id = self.meta_data["neuron_id"].to_numpy()
        positions = self.meta_data[["x", "y", "z"]].to_numpy()

        centre_point = np.mean(positions, axis=0)
        dist_to_centre = np.linalg.norm(positions - centre_point, axis=1)

        if distance_to_centre:
            print(f"Core filtering, maximal distance to centre {distance_to_centre} meters")
            distance_threshold = distance_to_centre
        else:
            assert population_fraction is not None, f"Either distance_to_centre or population_fraction must be set"

            print(f"Filtering simplices with at least one member in core "
                  f"(core fraction {population_fraction} of all neurons)")
            distance_threshold = np.percentile(dist_to_centre, population_fraction)

        centre_idx = np.where(dist_to_centre <= distance_threshold)[0]

        lookup_table = np.zeros((len(neuron_id),), dtype=bool)
        lookup_table[centre_idx] = True

        row_num = np.array([len(x) for x in self.data_file.values()])
        row_dim = np.array([self.get_dim_from_name(x) for x in self.data_file.keys()])

        # Dimensions might not be in order, reorder them
        sort_idx = np.argsort(row_dim)
        work_load = np.multiply(row_num[sort_idx], row_dim[sort_idx])
        work_load_done = 0
        total_work_load = np.sum(work_load)

        self.open_output()

        for dim, all_rows_on_file in self.data_file.items():

            all_rows = all_rows_on_file[()].copy()

            if self.check_if_exists(dim):
                print(f"{dim} already stored in {self.filtered_file_name}, skipping.")
                continue

            start_time = timeit.default_timer()

            print(f"Processing {dim} with {len(all_rows)}")
            keep_rows = np.zeros((len(all_rows,)), dtype=bool)
            filtered_data = []

            all_rows_flag = np.take(lookup_table, all_rows)
            keep_rows = np.where(np.sum(all_rows_flag, axis=1))[0]

            filtered_data = [all_rows[x] for x in keep_rows]
            print(f"Keeping {len(filtered_data)} out of {len(all_rows)} simplices")
            print(f"Time used for {dim}: {(timeit.default_timer() - start_time):0.1f} seconds")

            self.save_dim_data(dim_name=dim, dim_data=filtered_data)

            dim_value = self.get_dim_from_name(dim)
            work_load_done += work_load[dim_value-1]

            time_used = timeit.default_timer() - start_time
            fraction_done = work_load_done / total_work_load
            total_time_estimated = time_used / fraction_done
            time_left = total_time_estimated - time_used
            sys.stdout.write(f"\rTime used: {time_used:0.1f}s, estimated left {time_left:0.1f}s  "
                             f"({total_time_estimated:0.1f}s)  ")
            sys.stdout.flush()

        self.close()

        print(f"Total time used {(timeit.default_timer() - start_time):0.1f} seconds.")

    def open_output(self):

        if os.path.isfile(self.filtered_file_name):
            self.filtered_data_file = h5py.File(self.filtered_file_name, "a")
        else:
            self.filtered_data_file = h5py.File(self.filtered_file_name, "w")

            meta_info = {"neuron_id": self.meta_data["neuron_id"].to_numpy(),
                         "type": self.meta_data["neuron_type"].to_list(),
                         "name": self.meta_data["neuron_name"].to_list(),
                         "position": self.meta_data[["x", "y", "z"]].to_numpy(),
                         "morphology": self.meta_data["morphology"].to_list(),
                         "num_simplices": np.full((len(self.data_file)+1,), -1, dtype=int)}

            meta = self.filtered_data_file.create_group("meta")
            for key, value in meta_info.items():
                meta.create_dataset(key, data=value)

    @staticmethod
    def get_dim_from_name(dim_name):
        return int(dim_name[6:])

    def check_if_exists(self, dim_name):
        return dim_name in self.filtered_data_file and \
               self.filtered_data_file["meta/num_simplices"][self.get_dim_from_name(dim_name)] >= 0

    def save_dim_data(self, dim_name, dim_data):
        print(f"Writing {dim_name} ({len(dim_data)} rows)")
        self.filtered_data_file.create_dataset(dim_name, data=dim_data)
        self.filtered_data_file["meta/num_simplices"][self.get_dim_from_name(dim_name)] = len(dim_data)


def filter_flagser_cli():

    import argparse
    parser = argparse.ArgumentParser("Filter flagser")
    parser.add_argument("flagser_file", help="Flagser file")
    parser.add_argument("meta_data_file", help="Meta data file")
    parser.add_argument("output_file", help="Output file name")
    parser.add_argument("--percentile", help="Percentile of population kept (0-100)", type=int, default=None)
    parser.add_argument("--distance", help="Maximum distance to centre for core neurons (micrometers)", type=float, default=None)

    args = parser.parse_args()

    ff = FilterFlagserData(flagser_data_file_name=args.flagser_file,
                           network_meta_file_name=args.meta_data_file,
                           filtered_file_name=args.output_file)
    ff.load_data()
    ff.find_core_neurons(distance_to_centre=args.distance, population_fraction=args.percentile)


if __name__ == "__main__":
    filter_flagser_cli()